# -*- coding: utf-8 -*-

#######################################################################
# Copyright (C) 2022 Vinh Tran
#
#  This file is part of fDOG tool https://github.com/BIONF/fDOG
#
#  This script is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License <http://www.gnu.org/licenses/> for
#  more details
#
#  Contact: tran@bio.uni-frankfurt.de
#
#######################################################################

import sys
import os
from Bio import SeqIO

import fdog.libs.zzz as general_fn


##### FUNCTIONS FOR OUTPUT #####

def print_debug(debug, cat, msg):
    """ Print msg of a category in debug mode """
    if debug == True:
        if cat == '':
            print('#DEBUG#\t%s' % msg)
        else:
            print('#DEBUG#\t%s\n#DEBUG#\t%s' % (cat, msg))


def print_stdout(silentOff, msg):
    """ Print stdout """
    if silentOff == True:
        print(msg)


def check_output_exist(outfile, force, append):
    """ Check if outfile exists
    And decide depends on the choice of force or append option
    """
    if os.path.exists(outfile):
        if force == True:
            print('WARNING: %s will be deleted!' % outfile)
            os.remove(outfile)
        elif append == True:
            general_fn.check_file_exist(outfile)
            print('Result will be appended to %s!' % outfile)
        else:
            sys.exit(
                'WARNING: %s exists! ' % outfile
                + 'You still can run with --force or --append option')


def write_hamstr(hamstr_result, outpath, seqName, force, append):
    """ Write result of ortholog search into seqName.extended.fa """
    outfile = '%s/%s.extended.fa' % (outpath, seqName)
    outfile = os.path.abspath(outfile)
    check_output_exist(outfile, force, append)

    ### Write to output.extended.fa
    ortho_count = len(hamstr_result) - 1
    if append == True:
        if os.path.exists(outfile):
            old_result_tmp = SeqIO.to_dict((SeqIO.parse(open(outfile),'fasta')))
            old_result = {}
            for old_id in old_result_tmp:
                if not old_id in hamstr_result:
                    old_result[old_id] = str(old_result_tmp[old_id].seq)
            hamstr_result = {**hamstr_result, **old_result}

    with open(outfile, 'w') as out_file:
        for id, seq in hamstr_result.items():
            out_file.write('>%s\n%s\n' % (id, seq))
    return(
        'Found %s ortholog(s)!\nOutput file: %s' % (ortho_count, outfile))


def hamstr_2_profile(fa_file):
    """ Convert extended.fa file into phyloprofile file """
    if os.path.exists(fa_file):
        pp_file = fa_file.replace('.extended.fa', '.phyloprofile')
        fa = SeqIO.to_dict((SeqIO.parse(open(fa_file),'fasta')))
        with open(pp_file, 'w') as pp:
            pp.write('geneID\tncbiID\torthoID\n')
            for id in list(fa.keys()):
                tmp = id.split('|')
                pp.write('%s\tncbi%s\t%s\n' % (tmp[0], tmp[1].split('@')[1], id))


def add_all_taxa(pp_file, searchTaxa):
    """ Add all "missing" search taxa into phyloprofile file """
    missing_taxa = [] # missing_taxa = [ncbi_id]
    for taxon in searchTaxa.split(','):
        flag = general_fn.search_string_in_file(pp_file, taxon)
        if flag == 0:
            missing_taxa.append(taxon.split('@')[1])
    first_gene = ''
    if os.path.exists(pp_file):
        with open(pp_file, 'a') as pp:
            for line in general_fn.read_file(pp_file):
                if not line.startswith('geneID'):
                    if not first_gene:
                        first_gene = line.split('\t')[0]
                        for i in missing_taxa:
                            if len(line.split('\t')) == 5:
                                pp.write(f'{first_gene}\tncbi{i}\tfdogMA\tNA\tNA\n')
                            else:
                                pp.write(f'{first_gene}\tncbi{i}\tfdogMA\n')
                        break



# -*- coding: utf-8 -*-

#######################################################################
# Copyright (C) 2022 Vinh Tran
#
#  This script is used to setup fdog: install dependencies and
#  download pre-computed data
#
#  This script is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License <http://www.gnu.org/licenses/> for
#  more details
#
#  Contact: tran@bio.uni-frankfurt.de
#
#######################################################################

import sys
import os
import platform
import argparse
import subprocess
import shutil
from ete4 import NCBITaxa
from pathlib import Path
from importlib.metadata import version, PackageNotFoundError

import fdog.libs.zzz as general_fn
import fdog.libs.fas as fas_fn
import fdog.libs.alignment as align_fn


def check_conda_env():
    """ Return if a conda env is currently using """
    if 'CONDA_DEFAULT_ENV' in os.environ:
        if not os.environ['CONDA_DEFAULT_ENV'] == 'base':
            return(True)
    return(False)


def get_source_path():
    """ Get path of installed fDOG library """
    fdogPath = os.path.realpath(__file__).replace('/setupfDog.py','')
    return(fdogPath)


def get_data_path(fdogPath):
    """ Get path of fDOG data """
    pathconfigFile = fdogPath + '/bin/pathconfig.yml'
    if not os.path.exists(pathconfigFile):
        sys.exit('No pathconfig.yml found. Please run fdog.setup (https://github.com/BIONF/fDOG/wiki/Installation#setup-fdog).')
    else:
        cfg = general_fn.load_config(pathconfigFile)
        try:
            dataPath = cfg['datapath']
        except:
            try:
                corepath = cfg['corepath']
            except:
                pass
            try:
                searchpath = cfg['searchpath']
            except:
                pass
            try:
                annopath = cfg['annopath']
            except:
                pass
            dataPath = 'Core taxa: %s\nSearch taxa: %s\nAnnotations: %s' % (corepath, searchpath, annopath)
        return(dataPath)


def install_fas(woFAS, force):
    """ Install greedyFAS """
    if not woFAS:
        if force:
            try:
                subprocess.check_call([sys.executable, "-m", "pip", "uninstall", "greedyFAS", "-y"])
            except subprocess.CalledProcessError as e:
                # If pip returns a non-zero code during uninstallation
                print(f"ERROR: Failed to uninstall greedyFAS. Pip returned error code {e.returncode}.")
                sys.exit(1)
            except Exception as e:
                # If the package is not installed, pip returns a CalledProcessError with a specific message
                if "Skipping" in str(e) or "not installed" in str(e):
                    print(f"Package greedyFAS not found. Continuing...")
                else:
                    print(f"Unexpected error: {e}")
                    sys.exit(1)
        ### check if fas already installed
        try:
            fasVersion = subprocess.run(['fas.run --version'], shell = True, capture_output = True, check = True)
        except:
            print('=> greedyFAS (https://github.com/BIONF/FAS)')
            install_fas_cmd = 'pip install greedyFAS'
            try:
                subprocess.check_output([install_fas_cmd], shell = True, stderr = subprocess.STDOUT)
            except subprocess.CalledProcessError as e:
                sys.exit('\033[91mERROR: Problem with installing FAS! Please do it manually. See: https://github.com/BIONF/FAS!\033[0m')
        ### check if fas installed but not yet configured
        check_fas = fas_fn.check_fas_executable()


def install_fasta36(fdogPath, cwd):
    """ Install FASTA36 from source """
    try:
        subprocess.check_output(['which fasta36'], shell = True, stderr = subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
        print('=> FASTA36 (https://github.com/wrpearson/fasta36)')
        fasta36v = '36.3.8h_04-May-2020'
        fasta36url = 'https://github.com/wrpearson/fasta36/archive/refs/tags'
        fasta36file = 'v%s.tar.gz' % fasta36v
        if not os.path.exists('%s/bin/aligner/bin/fasta36' % fdogPath):
            if os.path.exists('%s/bin/aligner' % fdogPath):
                shutil.rmtree('%s/bin/aligner' % fdogPath)
            general_fn.download_file(fasta36url, fasta36file)
            shutil.unpack_archive(fasta36file, '%s/bin/' % fdogPath, 'gztar')
            os.remove(fasta36file)
            shutil.move('%s/bin/fasta36-%s' % (fdogPath, fasta36v), '%s/bin/aligner' % fdogPath)
            if 'Darwin' in platform.uname():
                make_cmd = 'make -f %s/bin/aligner/make/Makefile.os_x86_64 all' % fdogPath
            elif 'Linux' in platform.uname():
                make_cmd = 'make -f %s/bin/aligner/make/Makefile.linux64_sse2 all' % fdogPath
            else:
                sys.exit('\033[91mERROR: Cannot identify type of system (neither Linux nor Darwin/MacOS)\033[0m')
            try:
                print('Compiling fasta36. Please wait...')
                os.chdir('%s/bin/aligner/src' % fdogPath)
                subprocess.run(make_cmd, shell = True, check = True)
            except:
                sys.exit('\033[91mERROR: Cannot install FASTA36!\033[0m')
            os.chdir(cwd)
            if not os.path.exists('%s/bin/aligner/bin/fasta36' % fdogPath):
                sys.exit('\033[91mERROR: fasta36 not found! Please install it manually!\033[0m')
            else:
                print('FASTA36 installed at %s/bin/aligner/' % fdogPath)
        else:
            fasta36_path = align_fn.check_fasta36_executable(fdogPath)
            print('FASTA36 found at %s' % fasta36_path)


def check_dependencies(fdogPath):
    """ Check for missing dependencies
    Dependencies are specified in fdog/data/dependencies.txt file
    """
    missing = []
    dependencies = '%s/data/dependencies.txt' % fdogPath
    for tool in general_fn.read_file(dependencies):
        function = tool
        if tool == 'hmmer':
            function = 'hmmsearch'
        if tool == 'ncbi-blast+':
            function = 'blastp'
        try:
            subprocess.check_output(['which %s' % function], shell = True, stderr = subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            missing.append(tool)
    return(missing)


def download_data(dataPath, resetData):
    """ Downloade pre-calculated fDOG data """
    data_fdog_file = "data_fDOG_2024.tar.gz"
    checksum_data = "1748371655 621731824 $data_fdog_file"

    genome_path = '%s/searchTaxa_dir' % dataPath
    Path(genome_path).mkdir(parents = True, exist_ok = True)
    if len(general_fn.read_dir(genome_path)) < 1 or resetData:
        data_url = 'https://applbio.biologie.uni-frankfurt.de/download/hamstr_qfo'
        if os.path.exists(data_fdog_file) and resetData:
            os.remove(data_fdog_file)
        general_fn.download_file(data_url, data_fdog_file)
        # ####### temporary solution while the uni network does not work #########
        # wgetCmd = 'wget "https://www.dropbox.com/scl/fi/t2ln18k0jthc3y74s591q/data_HaMStR-2019c.tar.gz?rlkey=c66nc3eslqyn2a6k6ey4e678r&st=plzvbllv&dl=0"'
        # try:
        #     subprocess.run([wgetCmd], shell=True, check=True)
        #     shutil.move("data_HaMStR-2019c.tar.gz?rlkey=c66nc3eslqyn2a6k6ey4e678r&st=plzvbllv&dl=0", "data_HaMStR-2019c.tar.gz")
        # except:
        #     print('Problem occurred while download demo data from dropbox')
        # ########################################################################
        try:
            print('Extracting %s...' % data_fdog_file)
            shutil.unpack_archive(data_fdog_file, dataPath, 'gztar')
        except:
            sys.exit('\033[91mERROR: Cannot extract %s to %s!\033[0m' % (data_fdog_file, dataPath))
        if 'genome_dir' in general_fn.read_dir(dataPath):
            os.rename('%s/genome_dir' % dataPath, '%s/searchTaxa_dir' % dataPath)
            os.rename('%s/blast_dir' % dataPath, '%s/coreTaxa_dir' % dataPath)
            os.rename('%s/weight_dir' % dataPath, '%s/annotation_dir' % dataPath)
        if not 'assembly_path' in general_fn.read_dir(dataPath):
            os.makedirs(f'{dataPath}/assembly_path')
            shutil.copytree(f'{get_source_path()}/data/assembly_dir', f'{dataPath}/assembly_dir')
        # check downloaded data
        check_cmd = 'fdog.checkData -s %s/searchTaxa_dir -c %s/coreTaxa_dir -a %s/annotation_dir --reblast --ignoreAnno' % (dataPath, dataPath, dataPath)
        try:
            print('Checking downloaded data...')
            subprocess.run([check_cmd], stdout = subprocess.DEVNULL, check = True, shell = True)
        except:
            print('\033[96mWARNING: Problem with validating downloaded data. Please run fdog.checkData manually!\033[0m')
        os.remove(data_fdog_file)
        print('fDOG data downloaded and saved at %s' % dataPath)
    else:
        print('fDOG data found at %s' % dataPath)


def write_pathconfig(fdogPath, dataPath):
    """ Write data directories to pathconfig file """
    Path('%s/bin' % fdogPath).mkdir(parents = True, exist_ok = True)
    pathconfigFile = '%s/bin/pathconfig.yml' % fdogPath
    if os.path.exists(pathconfigFile):
        os.remove(pathconfigFile)
    with open(pathconfigFile, 'w') as cf:
        cf.write('datapath: \'%s\'\n' % dataPath)
        cf.write('corepath: \'%s/coreTaxa_dir\'\n' % dataPath)
        cf.write('searchpath: \'%s/searchTaxa_dir\'\n' % dataPath)
        cf.write('annopath: \'%s/annotation_dir\'\n' % dataPath)
        cf.write('assemblypath: \'%s/assembly_dir\'\n' % dataPath)


def main():
    fdog_version = version("fdog")
    parser = argparse.ArgumentParser(description='You are running fDOG version ' + str(fdog_version) + '.')
    required = parser.add_argument_group('required arguments')
    optional = parser.add_argument_group('optional arguments')
    required.add_argument('-d', '--dataPath', help='Output path for fDOG data', action='store', default='', required=True)
    optional.add_argument('--getSourcepath', help='Get path to installed fdog package', action='store_true', default=False)
    optional.add_argument('--getDatapath', help='Get fDOG default data path', action='store_true', default=False)
    optional.add_argument('--woFAS', help='Do not install FAS (https://github.com/BIONF/FAS)', action='store_true', default=False)
    optional.add_argument('--force', help='Force installing the dependencies', action='store_true', default=False)
    optional.add_argument('--resetData', help='Re-download precalculated fDOG data', action='store_true', default=False)

    ### parse arguments
    args = parser.parse_args()
    dataPath = os.path.abspath(args.dataPath)
    woFAS = args.woFAS
    force = args.force
    resetData = args.resetData


    ### get install path
    fdogPath = get_source_path()
    if args.getSourcepath:
        print(fdogPath)
        sys.exit()

    ### get data path
    if args.getDatapath:
        dataPath = get_data_path(fdogPath)
        print(dataPath)
        sys.exit()

    ### check if pathconfig file exists
    pathconfigFile = '%s/bin/pathconfig.yml' % fdogPath
    demo_cmd = 'fdog.run --seqFile infile.fa --jobName test --refspec HUMAN@9606@qfo24_02'
    if os.path.exists(pathconfigFile) and not force:
        check_fas = 1
        if not woFAS:
            check_fas = fas_fn.check_fas_executable()
        if check_fas == 1:
            print('fDOG seems to be ready to use!')
            print('You can test fDOG using the following command:\n%s' % demo_cmd)
        else:
            print('fDOG seems to be ready to use without FAS!')
            print('You can test fDOG using the following command:\n%s --fasOff' % demo_cmd)
        sys.exit()

    ### get ncbi taxonomy database for ete4
    print('*** Creating local NCBI taxonomy database...')
    ncbi = NCBITaxa()

    ### install dependencies
    print('*** Installing dependencies...')
    ## FAS
    if not woFAS:
        install_fas(woFAS, force)
    ## hmmer, blast+, clustalw, mafft, muscle
    missing_tools = check_dependencies(fdogPath)
    if force or len(missing_tools) > 0:
        if check_conda_env() == True:
            req_file = '%s/data/conda_requirements.yml' % fdogPath
            print('=> Dependencies in %s' % req_file)
            install_cmd = f'install -c conda-forge -c bioconda --file {req_file} -y'
            if shutil.which("micromamba"):
                install_cmd = f'micromamba {install_cmd}'
            elif shutil.which("mamba"):
                install_cmd = f'mamba {install_cmd}'
            else:
                install_cmd = f'conda {install_cmd}'
            if force:
                install_cmd += ' --force-reinstall'
            try:
                subprocess.call(install_cmd, shell=True)
            except:
                sys.exit(f'\033[91mERROR: Cannot install conda packages in {req_file}!\033[0m')
        else:
            try:
                subprocess.check_output(['which metaeuk'], shell = True, stderr = subprocess.STDOUT)
            except subprocess.CalledProcessError as e:
                missing_tools.append('metaeuk')
            install_cmd = 'sudo apt-get install -y -qq <tool>'
            sys.exit('\033[91mERROR: Please install these tools manually:\n%s\nusing the command: %s!\033[0m' % (', '.join(missing_tools), install_cmd))
    else:
        if check_conda_env() == True:
            print('=> Dependencies in %s/data/conda_requirements.yml already installed!' % fdogPath)
        else:
            print('=> Dependencies in %s/data/dependencies.txt already installed!' % fdogPath)
    check_dependencies(fdogPath)
    ## fasta36
    install_fasta36(fdogPath, os.getcwd())

    ### download pre-calculated data
    print('*** Downloading precalculated data...')
    ### Remove data if resetData is used
    if resetData:
        if os.path.exists(dataPath):
            print('fDOG data found in %s will be deleted! Enter to continue.' % dataPath)
            if general_fn.query_yes_no(''):
                shutil.rmtree(dataPath)

    Path(dataPath).mkdir(parents = True, exist_ok = True)
    download_data(dataPath, resetData)

    ### create pathconfig file
    write_pathconfig(fdogPath, dataPath)

    print('\033[96m==> FINISHED! fDOG data can be found at %s\033[0m' % dataPath)
    check_fas = fas_fn.check_fas_executable()
    if check_fas == 1:
        print('You can test fDOG using the following command:\n%s' % demo_cmd)
    else:
        print('You can test fDOG using the following command:\n%s --fasOff' % demo_cmd)

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

#######################################################################
# Copyright (C) 2022 Vinh Tran
#
#  This file is part of fDOG tool https://github.com/BIONF/fDOG
#
#  This script is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License <http://www.gnu.org/licenses/> for
#  more details
#
#  Contact: tran@bio.uni-frankfurt.de
#
#######################################################################

import sys
import os
from pathlib import Path
from os.path import isfile, join
import argparse
import subprocess
import re
import shutil
import multiprocessing as mp
from tqdm import tqdm
from ete4 import NCBITaxa
from importlib.metadata import version, PackageNotFoundError
import time

import fdog.libs.zzz as general_fn
import fdog.libs.preparation as prepare_fn
import fdog.libs.orthosearch as ortho_fn
import fdog.libs.corecompile as core_fn
import fdog.libs.fas as fas_fn
import fdog.libs.tree as tree_fn
import fdog.libs.output as output_fn



def get_sorted_files(directory):
    list = os.listdir(directory)
    pairs = []
    for file in list:
        location = os.path.join(directory, file)
        if isfile(location):
            size = os.path.getsize(location)
            pairs.append((size, file))
    pairs.sort(key=lambda s: s[0], reverse=True)
    return([x[1] for x in pairs])


def get_seed_name(seedFile):
    seqName = seedFile.rsplit('.', 1)[0]
    seqName = re.sub('[\|\.]', '_', seqName)
    return(seqName)


def create_core_jobs(args):
    (seed, core_options, other_options, inFol, outpath, silentOff) = args
    (coreArgs, orthoCoreArgs, otherCoreArgs) = core_options
    (refspec, reuseCore, forceCore, pathArgs, debug) = other_options
    (outpath, hmmpath, corepath, searchpath, annopath) = pathArgs
    seqFile = ('%s/%s' % (inFol, seed))
    seqName = get_seed_name(seed)
    if not os.path.exists('%s/core_orthologs/%s/hmm_dir/%s.hmm' % (outpath, seqName, seqName)) or forceCore == True:
        seed_id = prepare_fn.identify_seed_id(seqFile, refspec, corepath, debug, silentOff)
        if not seed_id == 'None':
            return([seqFile, seqName, refspec, seed_id,
                        reuseCore, forceCore, coreArgs, pathArgs, orthoCoreArgs,
                        otherCoreArgs, debug])
        else:
            print(f'WARNING: Cannot identify seed ID for {seqFile}!')


def compile_core(core_options, other_options, seeds, inFol, cpus, outpath, silentOff, jobName):
    core_compilation_jobs = []
    (coreArgs, orthoCoreArgs, otherCoreArgs) = core_options
    (cpus, debugCore, silentOff, noCleanup, force, append) = otherCoreArgs
    (refspec, reuseCore, forceCore, pathArgs, debug) = other_options
    (outpath, hmmpath, corepath, searchpath, annopath) = pathArgs
    pool = mp.Pool(cpus)
    begin = time.time()
    print('Preparing core compilation jobs...')
    core_job_file = '%s/%s_core_jobs.list' % (outpath, jobName)
    if os.path.exists(core_job_file) and os.stat(core_job_file).st_size > 0:
        print('... file contains jobs found (%s)' % core_job_file)
        core_compilation_jobs = general_fn.read_pyobj_file(core_job_file)
    else:
        prepare_jobs = []
        for seed in seeds:
            prepare_jobs.append([seed, core_options, other_options, inFol, outpath, silentOff])
        for _ in tqdm(pool.imap_unordered(create_core_jobs, prepare_jobs), total=len(prepare_jobs)):
            core_compilation_jobs.append(_)
        general_fn.save_pyobj(core_compilation_jobs, core_job_file)
    end = time.time()
    print('==> %s jobs will be run. Preparing finished in %s' % (len(core_compilation_jobs), '{:5.3f}s'.format(end - begin)))
    if len(core_compilation_jobs) > 0:
        core_runtime = []
        if debugCore == True or silentOff == True or len(core_compilation_jobs) == 1:
            for job in core_compilation_jobs:
                tmp_out = core_fn.run_compile_core(job)
                core_runtime.append(tmp_out)
        else:
            for _ in tqdm(pool.imap_unordered(core_fn.run_compile_core, core_compilation_jobs), total=len(core_compilation_jobs)):
                core_runtime.append(_)
        pool.close()
        pool.join()
        out = []
        for r in core_runtime:
            out.append('\t'.join(r))
        return(out)


def search_ortholog(options, seeds, inFol, outpath):
    (orthoArgs, otherArgs, pathArgs, refspec) = options
    (searchTaxa, cpus, debug, silentOff, noCleanup, force, append) = otherArgs
    ortho_runtime = []
    n = 1
    for seed in seeds:
        begin = time.time()
        seqFile = [inFol + '/' + seed]
        seqName = get_seed_name(seed)
        print('... %s (%s / %s)' % (seqName, n, len(seeds)))
        if not os.path.exists('%s/%s.extended.fa' % (outpath, seqName)) or force == True:
            hamstr_out = ortho_fn.run_hamstr([seqName, refspec, pathArgs, orthoArgs, otherArgs])
            output_fn.write_hamstr(hamstr_out, outpath, seqName, force, append)
            end = time.time()
            ortho_runtime.append('%s\t%s' % (seqName, '{:5.3f}s'.format(end - begin)))
        n += 1
    return(ortho_runtime)


def join_outputs(outpath, jobName, seeds, keep, silentOff):
    finalFa = '%s/%s.extended.fa' % (outpath, jobName)
    single_output_fol = '%s/%s_singleOutput' % (outpath, jobName)
    Path('%s/%s_singleOutput' % (outpath, jobName)).mkdir(parents=True, exist_ok=True)
    with open(finalFa,'wb') as wfd:
        for seed in seeds:
            seqName = get_seed_name(seed)
            resultFile = '%s/%s.extended.fa'  % (outpath, seqName)
            if silentOff == True:
                print(resultFile)
            if os.path.exists(resultFile):
                with open(resultFile,'rb') as fd:
                    shutil.copyfileobj(fd, wfd)
                if not os.path.exists('%s/%s.extended.fa' % (single_output_fol, seqName)):
                    shutil.move(resultFile, single_output_fol)
            if os.path.exists(outpath + '/' + seqName + '.fa'):
                os.remove(outpath + '/' + seqName + '.fa')
            if os.path.exists(os.getcwd() + '/' + seqName + '.fa'):
                os.remove(os.getcwd() + '/' + seqName + '.fa')
    if keep == True:
        try:
            print('Compressing single outputs...')
            shutil.make_archive(single_output_fol, 'gztar', single_output_fol)
        except:
            shutil.make_archive(single_output_fol, 'tar', single_output_fol)
    shutil.rmtree(single_output_fol)


def main():
    fdog_version = version("fdog")
    parser = argparse.ArgumentParser(description='You are running fDOG version ' + str(fdog_version) + '.',
                                     epilog="For more information on certain options, please refer to the wiki pages "
                                            "on github: https://github.com/BIONF/fDOG/wiki")
    required = parser.add_argument_group('Required arguments')
    required.add_argument('--seqFolder', help='Input folder containing the seed sequences (protein only) in fasta format',
                            action='store', default='', required=True)
    required.add_argument('--jobName', help='Job name. This will also be file name for the output',
                            action='store', default='', required=True)
    required.add_argument('--refspec', help='Reference taxon. It should be the species the seed sequence was derived from',
                            action='store', default='', required=True)

    optional_paths = parser.add_argument_group('Non-default directory options')
    optional_paths.add_argument('--outpath', help='Output directory', action='store', default='')
    optional_paths.add_argument('--hmmpath', help='Path for the core ortholog directory', action='store', default='')
    optional_paths.add_argument('--corepath', help='Path for the core taxa directory', action='store', default='')
    optional_paths.add_argument('--searchpath', help='Path for the search taxa directory', action='store', default='')
    optional_paths.add_argument('--annopath', help='Path for the pre-calculated feature annotion directory', action='store', default='')
    optional_paths.add_argument('--pathFile', help='Config file contains paths to data folder (in yaml format)', action='store', default='')

    core_options = parser.add_argument_group('Core compilation options')
    core_options.add_argument('--coreOnly', help='Compile only the core orthologs', action='store_true', default=False)
    core_options.add_argument('--reuseCore', help='Reuse existing core set of your sequence', action='store_true', default=False)
    core_options.add_argument('--minDist', help='Minimum systematic distance of primer taxa for the core set compilation. Default: genus',
                            choices=['species', 'genus', 'family', 'order', 'class', 'phylum', 'kingdom', 'superkingdom'],
                            action='store', default='genus')
    core_options.add_argument('--maxDist', help='Maximum systematic distance of primer taxa for the core set compilation. Default: kingdom',
                            choices=['species', 'genus', 'family', 'order', 'class', 'phylum', 'kingdom', 'superkingdom'],
                            action='store', default='kingdom')
    core_options.add_argument('--coreSize', help='Maximul number of orthologs in core set. Default: 6', action='store', default=6, type=int)
    core_options.add_argument('--coreTaxa', help='List of primer taxa that should exclusively be used for the core set compilation', action='store', default='')
    core_options.add_argument('--CorecheckCoorthologsOff', help='Turn off checking for co-ortholog of the reverse search during the core compilation',
                                action='store_true', default=False)
    core_options.add_argument('--coreRep', help='Obtain only the sequence being most similar to the corresponding sequence in the core set rather than all putative co-orthologs',
                                action='store_true', default=False)
    core_options.add_argument('--coreHitLimit', help='Number of hits of the initial pHMM based search that should be evaluated via a reverse search. Default: 3',
                                action='store', default=3, type=int)
    core_options.add_argument('--candScoreFactor', help='A multiplier used to set cutoff for candidate selection. A candidate alignment score must exceed the '
                                'this factor in order to be considered further. Default: 0.05',
                                action='store', default=0.05, type=float)
    core_options.add_argument('--alnStrategy', help='Specify the alignment strategy during core ortholog compilation. Default: local',
                                choices=['local', 'glocal', 'global'],
                                action='store', default='local')

    ortho_options = parser.add_argument_group('Ortholog search strategy options')
    ortho_options.add_argument('--searchTaxa', help='Specify file contains list of search taxa', action='store', default='')
    ortho_options.add_argument('--group', help='Allows to limit the search to a certain systematic group', action='store', default='')
    ortho_options.add_argument('--checkCoorthologsRefOff', help='Turn off checking for co-ortholog of the reverse search during the final ortholog search',
                                action='store_true', default=False)
    ortho_options.add_argument('--rbh', help='Requires a reciprocal best hit during the ortholog search to accept a new ortholog',
                                action='store_true', default=False)
    ortho_options.add_argument('--rep', help='Obtain only the sequence being most similar to the corresponding sequence in the core set rather than all putative co-orthologs',
                                action='store_true', default=False)
    ortho_options.add_argument('--lowComplexityFilter', help='Switch the low complexity filter for the blast search on. Default: False',
                                action='store_true', default=False)
    ortho_options.add_argument('--evalBlast', help='E-value cut-off for the Blast search. Default: 0.0001',
                                action='store', default=0.0001, type=float)
    ortho_options.add_argument('--evalHmmer', help='E-value cut-off for the HMM search. Default: 0.0001',
                                action='store', default=0.0001, type=float)
    ortho_options.add_argument('--hitLimit', help='number of hits of the initial pHMM based search that should be evaluated via a reverse search. Default: 10',
                                action='store', default=10, type=int)
    ortho_options.add_argument('--hmmScoreType', help='Choose type of hmm score (best domain or full sequence score) for validating HMM candidates (NOTE: applied also for the core compilation). Default: domain',
                                action='store', choices=['domain','sequence'], default='domain')
    ortho_options.add_argument('--scoreCutoff', help='Define the percent range of the hmms core of the best hit up to which a candidate of the hmmsearch will be subjected for further evaluation. Default: 10',
                                action='store', default=10, type=int)

    fas_options = parser.add_argument_group('FAS options')
    fas_options.add_argument('--coreFilter',
                                help='Specifiy mode for filtering core orthologs by FAS score. In \'relaxed\' mode candidates with insufficient FAS score will be disadvantaged. In \'strict\' mode candidates with insufficient FAS score will be deleted from the candidates list. The option \'--minScore\' specifies the cut-off of the FAS score.',
                                choices=['relaxed', 'strict'], action='store', default='')
    fas_options.add_argument('--minScore', help='Specify the threshold for coreFilter. Default: 0.75', action='store', default=0.75, type=float)
    fas_options.add_argument('--featureFile', help='File that contains tools used for FAS calculation', action='store', default='')

    addtionalIO = parser.add_argument_group('Other I/O options')
    addtionalIO.add_argument('--append', help='Append the output to existing output files', action='store_true', default=False)
    addtionalIO.add_argument('--force', help='Overwrite existing ortholog search output files', action='store_true', default=False)
    addtionalIO.add_argument('--forceCore', help='Overwrite existing core set of your sequence', action='store_true', default=False)
    addtionalIO.add_argument('--notAddingTaxa', help='Do not add all search taxa to phyloprofile output', action='store_true', default=False)
    addtionalIO.add_argument('--noCleanup', help='Temporary output will NOT be deleted. Default: False', action='store_true', default=False)
    addtionalIO.add_argument('--keep', help='Keep output of individual seed sequence. Default: False', action='store_true', default=False)
    addtionalIO.add_argument('--debug', help='Set this flag to obtain more detailed information about the ortholog search progress', action='store_true', default=False)
    addtionalIO.add_argument('--debugCore', help='Set this flag to obtain more detailed information about the core compilation actions', action='store_true', default=False)
    addtionalIO.add_argument('--silentOff', help='Show more output to terminal', action='store_true', default=False)

    optional = parser.add_argument_group('Other options')
    optional.add_argument('--fasOff', help='Turn OFF FAS support', action='store_true', default=False)
    optional.add_argument('--aligner', help='Choose between mafft-linsi or muscle for the multiple sequence alignment. DEFAULT: muscle',
        choices=['mafft-linsi', 'muscle'], action='store', default='muscle')
    optional.add_argument('--cpus', help='Determine the number of threads to be run in parallel. Default: 4', action='store', default=4, type=int)

    ### get arguments
    args = parser.parse_args()

    # required arguments
    inFol = os.path.abspath(args.seqFolder)
    jobName = args.jobName
    refspec = args.refspec

    # path arguments
    outpath = os.path.abspath(args.outpath)
    hmmpath = args.hmmpath
    corepath = args.corepath
    searchpath = args.searchpath
    annopath = args.annopath
    pathFile = args.pathFile

    # core compilation arguments
    coreOnly = args.coreOnly
    reuseCore = args.reuseCore
    minDist = args.minDist
    maxDist = args.maxDist
    coreSize = args.coreSize
    coreTaxa = args.coreTaxa
    if not coreTaxa == '':
        if os.path.exists(os.path.abspath(coreTaxa)):
            coreTaxa = os.path.abspath(coreTaxa)
    CorecheckCoorthologsOff = args.CorecheckCoorthologsOff
    coreRep = args.coreRep
    coreHitLimit = args.coreHitLimit
    distDeviation = max(round(args.candScoreFactor - 1, 2), 0) #args.distDeviation
    alnStrategy = args.alnStrategy

    # ortholog search arguments
    searchTaxa = args.searchTaxa
    if not searchTaxa == '':
        if os.path.exists(os.path.abspath(searchTaxa)):
            searchTaxa = os.path.abspath(searchTaxa)
    group = args.group
    if not group == '' and not searchTaxa == '':
        print('WARNING: Both --group and --searchTaxa are specified. Search taxa will be obtained only from %s!' % searchTaxa)
        group = ''
    checkCoorthologsRefOff = args.checkCoorthologsRefOff
    rbh = args.rbh
    rep = args.rep
    lowComplexityFilter = args.lowComplexityFilter
    evalBlast = args.evalBlast
    evalHmmer = args.evalHmmer
    hitLimit = args.hitLimit
    hmmScoreType = args.hmmScoreType
    scoreCutoff = args.scoreCutoff

    # fas arguments
    fasOff = args.fasOff
    coreFilter = args.coreFilter
    minScore = args.minScore
    featureFile = args.featureFile

    # other I/O arguments
    append = args.append
    force = args.force
    forceCore = args.forceCore
    noCleanup = args.noCleanup
    keep = args.keep
    debug = args.debug
    debugCore = args.debugCore
    silentOff = args.silentOff
    notAddingTaxa = args.notAddingTaxa

    # others
    aligner = args.aligner
    cpus = args.cpus
    if cpus > os.cpu_count():
        cpus = os.cpu_count()


    begin = time.time()
    ##### Check and group parameters
    print('##### PREPARING & CHECKING #####')
    (inFol, hmmpath, corepath, searchpath, annopath) = prepare_fn.check_input(
                    [inFol, refspec, outpath, hmmpath,
                    corepath, searchpath, annopath, pathFile])
    pathArgs = [outpath, hmmpath, corepath, searchpath, annopath]
    prepare_fn.check_blast_version(corepath, refspec)

    (invalid_minDist, invalid_maxDist, suggested_minRank, suggested_maxRank) = prepare_fn.check_ranks_core_taxa(corepath, refspec, minDist, maxDist)
    if len(invalid_minDist) > 0 or len(invalid_maxDist) > 0:
        if len(invalid_minDist) > 0:
            print(f'Invalid {minDist} (--minDist) for:\t{", ".join(invalid_minDist)}')
        if len(invalid_maxDist) > 0:
            print(f'Invalid {maxDist} (--maxDist) for:\t{", ".join(invalid_maxDist)}')
        if not minDist == "genus" and not maxDist == "kingdom":
            print(f'Please consider setting --minDist and --maxDist with these valid ranks:\n--minDist {suggested_minRank} --maxDist {suggested_maxRank}')
            sys.exit()
        else:
            print(f'WARNING: --minDist and --maxDist will be automatically changed to {suggested_minRank} and {suggested_maxRank}')
            minDist = suggested_minRank
            maxDist = suggested_maxRank

    if not fasOff:
        check_fas = fas_fn.check_fas_executable()
        if check_fas == 0:
            sys.exit('ERROR: FAS is not executable! You still can use fDOG with --fasOff!')
    check_time = time.time()
    print('==> Preparation finished in ' + '{:5.3f}s'.format(check_time - begin))

    ### START
    Path(outpath).mkdir(parents=True, exist_ok=True)
    multiLog = open(outpath + '/' + jobName + '_runtime.txt', "w")

    print('PID %s - Jobname %s'% (str(os.getpid()), jobName))
    multiLog.write('PID %s - Jobname %s\n'% (str(os.getpid()), jobName))
    seeds = get_sorted_files(inFol)
    end = time.time()
    print('==> Sort seed files finished in ' + '{:5.3f}s'.format(end - begin))

    ##### DO CORE COMPILATION
    if reuseCore == False:
        print('##### COMPILING CORE ORTHOLOG GROUPS #####')
        start = time.time()
        coreArgs = [minDist, maxDist, coreSize, coreTaxa, distDeviation,
                    alnStrategy, fasOff]
        orthoCoreArgs = [CorecheckCoorthologsOff, rbh, True, evalBlast/10,
                        lowComplexityFilter, evalHmmer/10, coreHitLimit, hmmScoreType,
                        scoreCutoff, aligner] # rep = True; e-value cutoff is 10x more stringent than from ortho search
        otherCoreArgs = [cpus, debugCore, silentOff, noCleanup, force, append]
        core_options = [coreArgs, orthoCoreArgs, otherCoreArgs]
        other_options = [refspec, reuseCore, forceCore, pathArgs, debug]
        core_runtime = compile_core(core_options, other_options, seeds, inFol, cpus, outpath, silentOff, jobName)
        end = time.time()
        multi_core_time = '{:5.3f}'.format(end-start)
        print('==> Core compilation finished in %ss\n' % multi_core_time)
        if len(core_runtime) > 1:
            multiLog.write('==> Core compilation finished in %ss\n%s\n' % (multi_core_time, '\n'.join(core_runtime)))
        else:
            multiLog.write('==> Core compilation finished in %ss\n' % multi_core_time)
    else:
        if not os.path.exists(hmmpath):
            sys.exit('--reuseCore was set, but no core orthologs found in %s! You could use --hmmpath to manually specify the core ortholog directory.' % outpath)


    ##### DO ORTHOLOG SEARCH USING HMM (HAMSTR)
    finalFa = '%s/%s.extended.fa' % (outpath, jobName)
    if not coreOnly:
        print('##### SEARCHING ORTHOLOGS #####')
        start = time.time()
        if not os.path.exists(finalFa) or force == True:
            ### get list of search taxa
            if not group == '':
                ### Check valid taxonomy group
                ncbi = NCBITaxa()
                group_id = ncbi.get_name_translator([group])
                if len(group_id) == 0:
                    exit('ERROR: Taxon group "%s" invalid!' % group)
                ### create taxonomy tree from list of search taxa
                searchTaxa = []
                tax_ids = ortho_fn.get_search_taxa_ids(searchTaxa, searchpath)

                for tax_id in tax_ids.keys():
                    check = tree_fn.check_taxon_group(group_id[group][0], tax_id, ncbi)
                    if check == True:
                        searchTaxa.append(tax_ids[tax_id])
                if debugCore:
                    print(searchTaxa)
                if len(searchTaxa) == 0:
                    exit('ERROR: No taxon found within %s taxonomy group!' % group)
                else:
                    searchTaxa = ','.join(searchTaxa)

            if len(searchTaxa) == '':
                searchTaxa = general_fn.read_dir(searchpath)
                searchTaxa = ','.join(searchTaxa)

            ### do ortholog search
            orthoArgs = [checkCoorthologsRefOff, rbh, rep, evalBlast,
                        lowComplexityFilter, evalHmmer, hitLimit, hmmScoreType, scoreCutoff, aligner]
            otherArgs = [searchTaxa, cpus, debug, silentOff, noCleanup, force, append]
            ortho_options = [orthoArgs, otherArgs, pathArgs, refspec]
            ortho_runtime = search_ortholog(ortho_options, seeds, inFol, outpath)
            end = time.time()
            multi_ortho_time = '{:5.3f}'.format(end-start)
            print('==> Ortholog search finished in %ss\n' % multi_ortho_time)
            multiLog.write('==> Ortholog search finished in %ss\n%s\n' % (multi_ortho_time, '\n'.join(ortho_runtime)))
            ### join output
            print('Joining single outputs...')
            start = time.time()
            join_outputs(outpath, jobName, seeds, keep, silentOff)
            end = time.time()
            print('==> Joining outputs finished in %ss\n' % '{:5.3f}'.format(end-start))

        ##### DO FINAL FAS CALCULATION
        if not fasOff:
            print('##### CALCULATING FAS SCORES #####')
            try:
                fasVersion = subprocess.run(['fas.run --version'], shell = True, capture_output = True, check = True)
            except:
                sys.exit('Problem with FAS! Please check https://github.com/BIONF/FAS or turn it off if not needed!')
            if os.path.exists(finalFa):
                start = time.time()
                fas_fn.calc_fas_multi(finalFa, outpath, annopath, cpus, featureFile)
                end = time.time()
                print('==> FAS calculation finished in ' + '{:5.3f}s'.format(end - start))
                multiLog.write('==> FAS calculation finished in ' + '{:5.3f}s'.format(end - start))
        else:
            output_fn.hamstr_2_profile(finalFa)

        ##### ADD ALL SEARCH TAXA INTO PhyloProfile OUTPUT
        if not notAddingTaxa:
            pp_file = f'{outpath}/{jobName}.phyloprofile'
            if not searchTaxa:
                tmp = general_fn.read_dir(searchpath)
                searchTaxa = ','.join(tmp)
            output_fn.add_all_taxa(pp_file, searchTaxa)

        end = time.time()
        print('==> fdogs.run finished in ' + '{:5.3f}s'.format(end - begin))

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

#######################################################################
# Copyright (C) 2022 Vinh Tran
#
#  This file is part of fDOG tool https://github.com/BIONF/fDOG
#
#  This script is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License <http://www.gnu.org/licenses/> for
#  more details
#
#  Contact: tran@bio.uni-frankfurt.de
#
#######################################################################

import sys
import os
import subprocess
import pyhmmer

import fdog.libs.output as output_fn

##### FUNCTIONS RELATED TO HMM #####

def create_hmm(aln_file, out_file):
    """ Create hmm file for an alinment file """
    hmmbuild_cmd = 'hmmbuild --amino %s %s' % (out_file, aln_file)
    try:
        subprocess.run(
            [hmmbuild_cmd], shell = True,
            stdout = open(os.devnull, 'wb'), check = True)
    except:
        sys.exit('ERROR: Error running hmmbuild %s' % hmmbuild_cmd)


def sort_hmm_hits(hmm_hits, hmm_score_type = 'domain', hitLimit = 10, scoreCutoff = 10, debug = False):
    """ Sort HMM hits
    Keep only n hits (n =< hitLimit), and hits that are not less than
    best_hit_domain_score * (100 - scoreCutoff) / 100
    Input hmm_hits is a pyhmmer.plan7.topHits object
    """
    best_score = -9999 # some "best" domains still have negative score!
    cutoff = ''
    score_dict = {}
    ori_hits = {}
    best_hit_score = -9999
    if hmm_score_type == 'domain':
        for hit in hmm_hits:
            ori_hits[hit.name.decode('ASCII')] = len(hit.domains)
            best_domain_score = -9999 #hit.domains[0].score
            best_domain_hit = ''
            if len(hit.domains) > 0:
                # get domain with best score for this hit
                for i in hit.domains:
                    if i.score > best_domain_score:
                        best_domain_score = i.score
                        best_domain_hit = i.hit.name.decode('ASCII')
                # add hit to score_dict with increasing domain score
                if best_domain_score > best_score:
                    best_score = best_domain_score
                cutoff = best_score/100*(100-scoreCutoff)
                if best_score < 0:
                    cutoff = best_score/100*(100+scoreCutoff)
                if best_domain_score >= cutoff:
                    if best_domain_score not in score_dict:
                        score_dict[best_domain_score] = [best_domain_hit]
                    else:
                        score_dict[best_domain_score].append(best_domain_hit)
    else:
        for hit in hmm_hits:
            ori_hits[hit.name.decode('ASCII')] = hit.score
            if hit.score > best_hit_score:
                # get hit with best score
                best_hit_score = hit.score
                best_hit = hit.name.decode('ASCII')
            # add to score_dict
            if best_hit_score > best_score:
                best_score = best_hit_score
            cutoff = best_score/100*(100-scoreCutoff)
            if best_score < 0:
                cutoff = best_score/100*(100+scoreCutoff)
            if hit.score >= cutoff:
                if hit.score not in score_dict:
                    score_dict[hit.score] = [hit.name.decode('ASCII')]
                else:
                    score_dict[hit.score].append(hit.name.decode('ASCII'))

    output_fn.print_debug(debug, 'All HMM hits', ori_hits)
    hmm_cand = {}
    n = 0
    score_dict = {
        key:val for key, val in score_dict.items() \
        if key >= cutoff
    }
    output_fn.print_debug(debug, 'Candidate HMM hits', score_dict)
    for score in sorted(score_dict, reverse = True):
        if n < hitLimit:
            for id in score_dict[score]:
                hmm_cand[id] = score
                n += 1
    return(hmm_cand)


def do_hmmsearch(
        hmm_file, search_fa, evalHmmer = 0.00001, scoreCutoff = 10,
        hitLimit = 10, hmm_score_type = 'domain', cpus = os.cpu_count(), debug = False):
    """ Perform hmmsearch for a hmm file vs a multiple fasta file
    Return a dictionary of hits and their e-value and bit-score
    Only "top" hits are returned. The cutoff is defined by
    max_score / 100 * (100 - scoreCutoff)
    By default, only hits that have at least 90% of the best bit score
    are considers
    """
    hmm_hits = {}
    with pyhmmer.easel.SequenceFile(search_fa, digital = True, alphabet = pyhmmer.easel.Alphabet.amino()) as seq_file:
        sequences = list(seq_file)
    with pyhmmer.plan7.HMMFile(hmm_file) as hmm_file:
        try:
            for hits in pyhmmer.hmmsearch(
                    hmm_file, sequences, E = evalHmmer, cpus = cpus):
                if len(hits) > 0:
                    hmm_hits = sort_hmm_hits(hits, hmm_score_type, hitLimit, scoreCutoff, debug)
        except :
            sys.exit(
                'ERROR: Error running hmmsearch for %s agains %s'
                % (hmm_file, search_fa))
    return(hmm_hits)



# -*- coding: utf-8 -*-

#######################################################################
# Copyright (C) 2022 Vinh Tran
#
#  This file is part of fDOG tool https://github.com/BIONF/fDOG
#
#  This script is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License <http://www.gnu.org/licenses/> for
#  more details
#
#  Contact: tran@bio.uni-frankfurt.de
#
#######################################################################

import sys
import os
from pathlib import Path
from Bio import SeqIO
import subprocess
import re
from datetime import datetime
from collections import OrderedDict

import fdog.libs.zzz as general_fn
import fdog.libs.blast as blast_fn
import fdog.libs.fasta as fasta_fn
import fdog.libs.tree as tree_fn

##### FUNCTIONS RELATED TO ADDING NEW TAXON TO FDOG DATABASE #####

def check_conflict_opts(replace, delete):
    """ Check if both replace and delete option are specified """
    if delete:
        if replace:
            sys.exit('*** ERROR: only one option can be choose between "--replace" and "--delete"')
    if replace:
        if delete:
            sys.exit('*** ERROR: only one option can be choose between "--replace" and "--delete"')


def get_paths(outPath, fdogPath, searchpath, corepath, annopath):
    """ Get path to searchTaxa_dir, coreTaxa_dir and annotation_dir """
    if outPath == '':
        pathconfigFile = fdogPath + '/bin/pathconfig.yml'
        if not os.path.exists(pathconfigFile):
            sys.exit('No pathconfig.yml found. Please run fdog.setup (https://github.com/BIONF/fDOG/wiki/Installation#setup-fdog).')
        cfg = general_fn.load_config(pathconfigFile)
        try:
            outPath = cfg['dataPath']
        except:
            try:
                corepath = cfg['corepath']
            except:
                pass
            try:
                searchpath = cfg['searchpath']
            except:
                pass
            try:
                annopath = cfg['annopath']
            except:
                pass

    outPath = os.path.abspath(outPath)
    if not searchpath:
        searchpath = outPath + '/searchTaxa_dir/'
    searchpath = os.path.abspath(searchpath)
    if not corepath:
        corepath = outPath + '/coreTaxa_dir/'
    corepath = os.path.abspath(corepath)
    if not annopath:
        annopath = outPath + '/annotation_dir/'
    annopath = os.path.abspath(annopath)
    return(outPath, searchpath, corepath, annopath)

def create_folders(searchpath, corepath, annopath, spec_name, coreTaxa, noAnno):
    """ Create searchTaxa_dir, coreTaxa_dir and annotation_dir in output folder """
    Path(searchpath).mkdir(parents = True, exist_ok = True)
    genome_path = '%s/%s' %  (searchpath, spec_name)
    Path(genome_path).mkdir(parents = True, exist_ok = True)
    if coreTaxa:
        Path(corepath).mkdir(parents = True, exist_ok = True)
    if not noAnno:
        Path(annopath).mkdir(parents = True, exist_ok = True)
    return(genome_path)


def generate_spec_name(tax_id, name, ver):
    """ Create species name with the format <ABBR_NAME>@<NCBI_TAXID>@<VER> """
    if name == "":
        ncbi_name = tree_fn.check_tax_id(tax_id)
        name = tree_fn.abbr_ncbi_name(ncbi_name)
    return(name+'@'+tax_id+'@'+ver)


def create_genome(args):
    """ Create fa and fai in searchTaxa_dir """
    (faIn, genome_path, spec_name, force, replace, delete) = args
    ### load fasta seq
    in_seq = SeqIO.to_dict((SeqIO.parse(open(faIn), 'fasta')))
    if not os.path.exists(genome_path):
        Path(genome_path).mkdir(parents = True, exist_ok = True)
    genome_file = '%s/%s.fa' % (genome_path, spec_name)
    if (not os.path.exists(os.path.abspath(genome_file))) or (os.stat(genome_file).st_size == 0) or force:
        f = open(genome_file, 'w')
        pipe = 0
        long_id = 0
        mod_id_index = 0
        id_dict = {} # id_dict[ori_id] = mod_id
        for id in in_seq:
            ori_id = id
            seq = str(in_seq[id].seq)
            ### check if ID contains empty char or pipe
            if ' ' in id:
                sys.exit('\033[91mERROR: Sequence IDs (e.g. %s) must not contain space(s)!\033[0m' % id)
            else:
                if '|' in id:
                    tmp = re.split('[_|]', id)
                    tmp = list(OrderedDict.fromkeys(tmp))
                    pipe = 1
                    id = '_'.join(tmp)
                    if not ori_id in id_dict:
                        id_dict[ori_id] = id
            ### check if id longer than 20 character
            if len(id) > 20:
                long_id = 1
                mod_id_index = mod_id_index + 1
                id = '%s_%s' % (spec_name.split('@')[1], mod_id_index)
                id_dict[ori_id] = id
            ### check if seq contains special characters
            if seq[-1] == '*':
                seq = seq[:-1]
            specialChr = 'no'
            if any(c for c in seq if not c.isalpha()):
                specialChr = 'yes'
            if specialChr == 'yes':
                if replace or delete:
                    if replace:
                        seq = re.sub('[^a-zA-Z]', 'X', seq)
                    if delete:
                        seq = re.sub('[^a-zA-Z]', '', seq)
                else:
                    sys.exit('\033[91mERROR: %s sequence contains special character!\033[0m\nYou can use --replace or --delete to solve it.' % (id))
            f.write('>%s\n%s\n' % (id, seq))
        f.close()
        ### create index file
        fasta_fn.read_fasta(genome_file)
        ### write .checked file
        cf = open(genome_file+'.checked', 'w')
        cf.write(str(datetime.now()))
        cf.close()
        ### write ID mapping file and give warning if ID changed
        if len(id_dict) > 0:
            mapping_file = '%s.mapping' % genome_file
            with open(mapping_file, 'w') as mp:
                for id in in_seq:
                    if id in id_dict:
                        mp.write(f'{id}\t{id_dict[id]}\n')
                    else:
                        mp.write(f'{id}\t{id}\n')
            if pipe == 1:
                print('\033[94mWARNING: Sequence IDs contain pipe(s). They will be replaced by "_"!\033[0m')
            if long_id == 'yes':
                print('\033[94mWARNING: Some headers longer than 80 characters have been automatically shortened.\033[0m')
            print('\033[94mPlease check the %s file for details!\033[0m' % mapping_file)
    else:
        print(genome_path + '/' + spec_name + '.fa already exists!')
    return(genome_file)


def create_blastdb(args):
    """ Create blastdb for a given fasta genome_file """
    (searchpath, corepath, outPath, spec_name, genome_file, force, silent) = args
    blast_path = '%s/%s' % (corepath, spec_name)
    if (not os.path.exists(os.path.abspath('%s/%s.phr' % (blast_path, spec_name)))) or force:
        blast_fn.make_blastdb([spec_name, genome_file, outPath, corepath, searchpath, silent])
        ### make symlink to fasta files
        fa_in_genome = "%s/%s/%s.fa" % (searchpath, spec_name, spec_name)
        fai_in_genome = "%s/%s/%s.fa.fai" % (searchpath, spec_name, spec_name)
        fa_in_blast = "%s/%s.fa" % (blast_path, spec_name)
        fai_in_blast = "%s/%s.fa.fai" % (blast_path, spec_name)
        if not os.path.exists(fa_in_blast):
            os.symlink(fa_in_genome, fa_in_blast)
        if not os.path.exists(fai_in_blast):
            os.symlink(fai_in_genome, fai_in_blast)
    else:
        print('Blast DB already exists!')


def create_annoFile(annopath, genome_file, cpus, force):
    """ Create annotation json for a given genome_file """
    annoCmd = 'fas.doAnno -i %s -o %s --cpus %s' % (genome_file, annopath, cpus)
    if force:
        annoCmd = annoCmd + " --force"
    try:
        subprocess.call([annoCmd], shell = True)
    except:
        print('\033[91mERROR: Problem with running fas.doAnno. You can check it with this command:\n%s\033[0m' % annoCmd)



import fdog.libs.addtaxon as addTaxon_fn
import fdog.libs.tree as tree_fn
import fdog.libs.zzz as general_fn
import sys
import os
import argparse

def check_fasta(file):
    nHeader = general_fn.count_line(file, '>', True)
    nSeq = general_fn.count_line(file, '>', False)
    if not nHeader == nSeq:
        return(1)
    nPipe = general_fn.count_line(file, '|', True)
    if nPipe > 0:
        return(1)
    return(0)

def check_path(path):
    if not os.path.exists(path):
        return False
    else:
        if os.path.isfile(path):
            return "File"
        else:
            return "Path"

def parse_file(path):
    file = open(path, "r")
    lines = file.readlines()
    id_dict = {}
    for line in lines:
        line = line.rstrip()
        ncbi, name = line.split("\t")
        id_dict[ncbi] = name
    file.close()
    return id_dict


def main():
    print("#################################")
    #################### handle user input #####################################
    version = '0.0.3'
    ################### initialize parser ######################################
    parser = argparse.ArgumentParser(description='You are running fdog.addAssembly version ' + str(version) + '.')
    ################## required arguments ######################################
    required = parser.add_argument_group('Required arguments')
    required.add_argument('--fasta', help='Path to fasta file or folder', action='store', default='', required=True)
    required.add_argument('--out', help='Path to output folder.', action='store', default='', required=True)
    required.add_argument('--ncbi', help='NCBI ID of species or a mapping file (tab separated) containing the NCBI ID and the corresponding file name placed in the folder given by --fasta. ', action='store', default='', required=True)
    required.add_argument('--ver', help='Version', action='store', default='', required=True)
    optional = parser.add_argument_group('Optional arguments')
    optional.add_argument('--link', help='links fasta files instead of copying them', action='store_true', default = False)

    args = parser.parse_args()
    fasta = args.fasta
    if check_path(fasta) == False:
        print("%s does not exists. Exiting ..."%(fasta))
        sys.exit()
    else:
        format = check_path(fasta)
    out_folder = args.out
    out_folder = os.path.abspath(out_folder) + '/'
    os.system('mkdir %s >/dev/null 2>&1' % (out_folder))
    ncbi = args.ncbi
    ver = args.ver
    ln = args.link
    id_dict = {}

    if check_path(ncbi) == False:
        if ncbi.isdigit() and format == "File":
            id_dict[ncbi] = fasta
        else:
            print("%s is no file or digit. Exiting ..."%(ncbi))
            sys.exit()
    elif check_path(ncbi) == "File":
        print("Parsing mapping file ...")
        id_dict = parse_file(ncbi)
        print("... done")
    else:
        print("%s is no file or digit. Exiting ..."%(ncbi))
        sys.exit()
    #print(format)
    #print(fasta)
    if format == "File":
        fa = id_dict[ncbi]
        if check_fasta(fa):
            name = addTaxon_fn.generate_spec_name(ncbi, "", ver)
            if ln == False:
                assembly_folder = out_folder + name
                os.system('mkdir %s >/dev/null 2>&1' % (assembly_folder))
                os.system("cp %s %s/%s.fa" %(fa, assembly_folder, name))
            else:
                assembly_folder = out_folder + name
                os.system('mkdir %s >/dev/null 2>&1' % (assembly_folder))
                os.system("ln %s %s/%s.fa" %(fa, assembly_folder, name))
        else:
            print("%s Fasta format not valid or header includes |"%(fa))
    
    else:
        for sp in id_dict:
            print("Adding species %s"%(sp))
            #print(id_dict)
            fa = id_dict[sp]
            fasta = os.path.abspath(fasta) + '/'
            #print(fa)
            #print(fasta)
            fasta_path = fasta + fa
            if check_fasta(fasta_path):
                name = addTaxon_fn.generate_spec_name(sp, "", ver)
                if ln == False:
                    assembly_folder = out_folder + name
                    os.system('mkdir %s >/dev/null 2>&1' % (assembly_folder))
                    os.system("cp %s %s/%s.fa" %(fasta_path, assembly_folder, name))
                else:
                    assembly_folder = out_folder + name
                    os.system('mkdir %s >/dev/null 2>&1' % (assembly_folder))
                    os.system("ln -s %s %s/%s.fa" %(fasta_path, assembly_folder, name))
            else:
                print("%s Fasta format not valid or header includes |"%(fasta_path))

    print("DONE, files can be found: %s"%(out_folder))

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

#######################################################################
# Copyright (C) 2021 Hannah Muelbaier
#
#  This script is used to prepare the core group used as input for fDOG-Assembly from a fasta file of an ortholog group.
#
#  This script is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License <http://www.gnu.org/licenses/> for
#  more details
#
#  Contact: hannah.muelbaier@gmail.com
#
#######################################################################

############################ imports ##################################
import os
import os.path
import sys
import argparse
import fdog.libs.alignment as align_fn
import fdog.libs.zzz as general_fn

def check_fasta(file):
    nHeader = general_fn.count_line(file, '>', True)
    nSeq = general_fn.count_line(file, '>', False)
    if not nHeader == nSeq:
        return(1)
    return(0)

def make_single_line_fasta(input, gene, out_folder):
    output = out_folder + gene + ".fa"
    with open(input, 'r') as f_input, open(output, 'w') as f_output:
        block = []
        for line in f_input:
            if line.startswith('>'):
                if block:
                    f_output.write(''.join(block) + '\n')
                    block = []
                f_output.write(line)
            else:
                block.append(line.strip())

        if block:
            f_output.write(''.join(block) + '\n')
    return (output)

def makeMSA(out_folder, gene, fasta_file):
    aln_file = out_folder + gene + '.aln'
    if align_fn.get_muscle_version('muscle') == 'v3':
        os.system('muscle -quiet -in %s -out %s' % (fasta_file, aln_file))
    else:
        os.system('muscle -align %s -output %s' % (fasta_file, aln_file))
    return aln_file

def makeHMM(out_folder, gene, aln_file):
    hmm_dir = out_folder + 'hmm_dir'
    os.system('mkdir %s >/dev/null 2>&1' % (hmm_dir))
    out_file = '%s/%s.hmm' % (hmm_dir, gene)
    hmmbuild_cmd = 'hmmbuild --amino %s %s' % (out_file, aln_file)
    os.system(hmmbuild_cmd)
    return out_file


def main():

    #################### handle user input #####################################
    version = '0.0.1'
    ################### initialize parser ######################################
    parser = argparse.ArgumentParser(description='You are running fdog.addCoreGroup version ' + str(version) + '.')
    ################## required arguments ######################################
    required = parser.add_argument_group('Required arguments')
    required.add_argument('--fasta', help='Path to fasta file of ortholog group.', action='store', default='', required=True)
    required.add_argument('--out', help='Path to output folder.', action='store', default='', required=True)
    required.add_argument('--geneName', help='Core group name', action='store', default='', required=True)
    args = parser.parse_args()

    fasta_file_input = args.fasta
    out_folder = args.out
    gene = args.geneName


    out_folder = out_folder + '/' + gene + '/'
    os.system('mkdir %s >/dev/null 2>&1' % (out_folder))

    if check_fasta(fasta_file_input) == 1:
        fasta_file = make_single_line_fasta(fasta_file_input, gene, out_folder)
    else:
        fasta_file = out_folder + gene + '.fa'
        os.system('cp ' + fasta_file_input + ' ' + fasta_file)

    aln_file = makeMSA(out_folder, gene, fasta_file)
    hmm_file = makeHMM(out_folder, gene, aln_file)

    print('Core group located at %s. Fasta file: %s; MSA: %s; HMM: %s' % (out_folder, fasta_file, aln_file, hmm_file))

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

#######################################################################
# Copyright (C) 2021 Hannah Muelbaier
#
#  This script is used to run fDOG-Assembly which performs targeted ortholog
#  searches on genome assemblies
#
#  This script is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License <http://www.gnu.org/licenses/> for
#  more details
#
#  Contact: hannah.muelbaier@gmail.com
#
#######################################################################

############################ imports ###########################################
import os
import os.path
import sys
from Bio import SeqIO
from Bio.Phylo.TreeConstruction import DistanceCalculator
from Bio import AlignIO
import argparse
import yaml
import subprocess
import time
import shutil
import multiprocessing as mp
import fdog.libs.alignment as align_fn
import fdog.libs.zzz as general_fn
import fdog.libs.fas as fas_fn
from tqdm import tqdm
from pathlib import Path
import pandas as pd

########################### functions ##########################################
def check_path(path, exit=True):
    if not os.path.exists(path) and exit == True:
        print(path + " does not exist. Exciting ...")
        sys.exit()
    elif not os.path.exists(path) and exit == False:
        return 1
    else:
        return 0

def check_ref_spec(species_list, fasta_file):
    """ Checks if reference species is part of the input ortholog group
    """
    species_file = set()
    with open(fasta_file,"r") as lines:
        for line in lines:
            if line[0] == ">":
                species = line.split("|")[1]
                species_file.add(species)
    for species in species_list:
        if species in species_file:
            return species
    print("Reference species is not part of the ortholog group. Exciting ...")
    sys.exit()


def starting_subprocess(cmd, mode, time_out = None):

    try:
        if mode == 'debug':
            result = subprocess.run(cmd, shell=True, timeout = time_out)
            return result
        elif mode == 'silent':
            result = subprocess.run(cmd, stdout = subprocess.PIPE, stderr = subprocess.PIPE, shell=True, timeout = time_out)
            return result
        elif mode == 'normal':
            result = subprocess.run(cmd, stdout = subprocess.PIPE, shell=True, timeout = time_out)
            return result
    except subprocess.TimeoutExpired:
        return 1

def merge(blast_results, insert_length):
    #merging overlapping and contigous candidate regions
    #format dictionary: {node_name: [(<start>,<send>,evalue, <qstart>,<qend>,<strand>, <score>)]}
    number_regions = 0
    insert_length = int(insert_length)
    score_list = []
    for key in blast_results:
        locations = blast_results[key]
        locations = sorted(locations, key = lambda x: int(x[3]))
        size_list = len(locations)
        j = 0
        while j < size_list-1:
            i = j + 1
            while i < size_list:
                if ((locations[j][0] < locations[i][0]) and (locations[j][1] > locations[i][0]) and (locations[j][5] == locations[i][5]) and (locations[i][5] == '+')):
                    #merge overlapping regions plus strand
                    locations[j][1] = max(locations[j][1], locations[i][1])
                    locations[j][2] = min(locations[j][2], locations[i][2])
                    locations[j][4] = max(locations[j][4], locations[i][4])
                    locations[j][6] = max(locations[j][6], locations[i][6])
                    locations.pop(i)
                    size_list -= 1
                    i -= 1
                elif ((locations[j][1] > locations[i][1]) and (locations[j][0] < locations[i][1]) and (locations[j][5] == locations[i][5]) and (locations[i][5] == '-')):
                    #merge overlapping regions minus strand
                    locations[j][0] = min(locations[j][0], locations[i][0])
                    locations[j][2] = min(locations[j][2], locations[i][2])
                    locations[j][4] = max(locations[j][4], locations[i][4])
                    locations[j][6] = max(locations[j][6], locations[i][6])
                    locations.pop(i)
                    size_list -= 1
                    i -= 1
                elif ((locations[j][0] < locations[i][0]) and (locations[i][0] - locations[j][1] <= 2*insert_length) and (locations[j][5] == locations[i][5]) and (locations[i][5] == '+')):
                    #merging consecutive regions, the distance between booth is not longer than a cutoff, plus strand
                    locations[j][1] = max(locations[j][1], locations[i][1])
                    locations[j][2] = min(locations[j][2], locations[i][2])
                    locations[j][4] = max(locations[j][4], locations[i][4])
                    locations[j][6] = max(locations[j][6], locations[i][6])
                    locations.pop(i)
                    size_list -= 1
                    i -=1
                elif ((locations[j][1] > locations[i][1]) and (locations[j][0] - locations[i][1] <= 2* insert_length) and (locations[j][5] == locations[i][5]) and (locations[i][5] == '-')):
                    #merging consecutive regions, the distance between booth is not longer than a cutoff, minus strand
                    locations[j][0] = min(locations[j][0], locations[i][0])
                    locations[j][2] = min(locations[j][2], locations[i][2])
                    locations[j][4] = max(locations[j][4], locations[i][4])
                    locations[j][6] = max(locations[j][6], locations[i][6])
                    locations.pop(i)
                    size_list -= 1
                    i -=1
                i += 1
            j += 1

        for entry in locations:
            score_list.append(entry[6])
        number_regions += len(locations)
        blast_results[key] = locations

    return blast_results, number_regions, score_list

def parse_blast(line, blast_results, cutoff):
    # format blast line:  <contig> <sstart> <send> <evalue> <qstart> <qend> <score>
    # format dictionary: {node_name: [(<start>,<send>,evalue, <qstart>,<qend>,<strand>, <score>)]}
    line = line.replace("\n", "")
    line_info = line.split("\t")
    evalue = float(line_info[3])
    #cut off
    if evalue > cutoff:
        return blast_results, evalue
    #add region to dictionary
    else:
        node_name, sstart, send, qstart, qend, score = line_info[0], int(line_info[1]), int(line_info[2]), int(line_info[4]), int(line_info[5]), int(line_info[6])
        split = node_name.split("|")
        # finding out on which strand tBLASTn found a hit
        if sstart < send:
            strand = "+"
        else:
            sstart = int(line_info[2])
            send = int(line_info[1])
            strand = "-"
        #creating a dictionary that inlcudes every tBLASTn that is better as the evalue cut-off
        if len(split) > 1:
            node_name = split[1]
        if node_name in blast_results:
            list = blast_results[node_name]
            list.append([int(sstart),int(send), evalue, int(qstart), int(qend), strand, score])
            blast_results[node_name] = list
        else:
            blast_results[node_name] = [[int(sstart),int(send), evalue, int(qstart), int(qend), strand, score]]

    return blast_results, evalue

def get_x_results(blast_dic, x, score_list):

    new_dic = {}
    score_list.sort(reverse=True)
    min = score_list[x - 1]
    number_regions = 0

    for key in blast_dic:
        key_list = []
        entries = blast_dic[key]
        for i in entries:
            if i[6] >= min:
                key_list.append(i)
        if key_list != []:
            new_dic[key] = key_list
            number_regions += len(key_list)
    return new_dic, number_regions

def candidate_regions(intron_length, cutoff_evalue, tmp_path, x = 10):
    ###################### extracting candidate regions ########################
    # info about output blast http://www.metagenomics.wiki/tools/blast/blastn-output-format-6
    blast_file = open(tmp_path + "/blast_results.out", "r")
    evalue = 0
    blast_results = {}
    #parsing blast output
    while True:
        line = blast_file.readline()
        #end of file is reached
        if not line:
            break
        #parsing blast output
        blast_results, evalue = parse_blast(line, blast_results, cutoff_evalue)

    if blast_results == {}:
        blast_file.close()
        return 0,0
    else:
        candidate_regions, number_regions, score_list = merge(blast_results, intron_length)
        blast_file.close()
        if number_regions > x:
            candidate_regions, number_regions = get_x_results(candidate_regions, x, score_list)
        return candidate_regions, number_regions

def extract_seq(region_dic, path, tmp_path, mode):

    for key in region_dic:
        #print("blastdbcmd -db " + path + " -dbtype 'nucl' -entry " + key + " -out tmp/" + key + ".fasta -outfmt %f")
        cmd = "blastdbcmd -db " + path + " -dbtype 'nucl' -entry " + key + " -out " + tmp_path + key + ".fasta -outfmt %f"
        starting_subprocess(cmd, mode)

def extract_sequence_from_to(name, file, start, end):
    #print(name)
    out = name + ".fasta"
    if int(start) < 0:
        start = 0
    with open(out,"w") as f:
        for seq_record in SeqIO.parse(file, "fasta"):
                f.write(">" + str(seq_record.id) + "\n")
                sequence_length = len(seq_record.seq)
                if int(end) > sequence_length:
                    end = sequence_length
                #for testing only
                #start = 0
                #end = len(seq_record.seq)
                f.write(str(seq_record.seq[int(start):int(end)]) + "\n")

    return out, start, end

def augustus_ppx(regions, candidatesOutFile, length_extension, profile_path, augustus_ref_species, ass_name, group, tmp_path, mode):
    """Gene prediction with software Augustus for all candidate regions. The resulting AS sequences will be written in a tmp file."""
    output = open(candidatesOutFile, "w")
    region = open(candidatesOutFile.replace(".candidates.fa", ".regions.txt"), "w")
    region.write("Contig/scaffold" + "\t" + "start" + "\t" + "end" + "\n")
    for key in regions:
        locations = regions[key]
        counter = 0
        for i in locations:
            # some variables
            counter += 1
            start = str(i[0] - length_extension)
            end = str(i[1] + length_extension)
            name = key + "_" + str(counter)
            # augutus call
            cmd = "augustus --protein=1 --gff3=on --proteinprofile=" + profile_path + " --predictionStart=" + start + " --predictionEnd=" + end + " --species=" + augustus_ref_species + " " + tmp_path + key + ".fasta > " + tmp_path + name + ".gff"
            #print(cmd)
            starting_subprocess(cmd, 'normal')
            # transfer augustus output to AS sequence
            #print(tmp_path)
            #print(key)
            cmd = "getAnnoFasta.pl --seqfile=" + tmp_path + key + ".fasta " + tmp_path + name + ".gff"
            #print(cmd)
            starting_subprocess(cmd, mode)
            #write region in region file
            region.write(key + "\t" + str(start) + "\t" + str(end) + "\n")
            # parsing header and sequences
            try:
                sequence_file = open(tmp_path + name + ".aa", "r")
                lines = sequence_file.readlines()
                for line in lines:
                    if line[0] == ">":
                        id = line.replace(">", "")
                        header = ">" + group + "|" + ass_name + "|" + name + "_" + id
                        output.write(header)
                    else:
                        output.write(line)
                sequence_file.close()
            except FileNotFoundError:
                pass
                #print("No gene found in region with ID" + name + " in species " + ass_name + " , continuing with next region")
    output.close()
    region.close()

def metaeuk_single(regions, candidatesOutFile, length_extension, ass_name, group, tmp_path, mode, db):
    output = open(candidatesOutFile, "w")
    region = open(candidatesOutFile.replace(".candidates.fa", ".regions.txt"), "w")
    region.write("Contig/scaffold" + "\t" + "start" + "\t" + "end" + "\n")

    for key in regions:
        locations = regions[key]
        counter = 0
        for i in locations:
            #some variables
            counter += 1
            start = str(i[0] - length_extension)
            end = str(i[1] + length_extension)
            name = key + "_" + str(counter)
            file, start, end = extract_sequence_from_to(tmp_path + name, tmp_path + key + ".fasta", start, end)
            region.write(key + "\t" + str(start) + "\t" + str(end) + "\n")
            #metaeuk call sensitive
            #cmd = "metaeuk easy-predict " + file + " " + db + " " + tmp_path + name + " " + tmp_path + "/metaeuk --min-exon-aa 5 --max-overlap 5 --min-intron 1 --overlap 1 --remove-tmp-files -s 6"
            #print(cmd)
            cmd = "metaeuk easy-predict " + file + " " + db + " " + tmp_path + name + " " + tmp_path + "/metaeuk --max-intron 130000 --max-seq-len 160000 --min-exon-aa 15 --max-overlap 15 --min-intron 5 --overlap 1 -s 4.5 --remove-tmp-files"
            # other parameteres used by BUSCO with metazoa set--max-intron 130000 --max-seq-len 160000 --min-exon-aa 5 --max-overlap 5 --min-intron 1 --overlap 1
            starting_subprocess(cmd, mode)
            # parsing header and sequences
            try:
                sequence_file = open(tmp_path + name + ".fas", "r")
                lines = sequence_file.readlines()
                #print(lines)
                id = 0
                for line in lines:
                    if line[0] == ">":
                        id += 1
                        header = ">" + group + "|" + ass_name + "|" + name + "_" + str(id) + "\n"
                        output.write(header)
                    else:
                        output.write(line)
                sequence_file.close()

                gff_file = open(tmp_path + name + ".gff", "r")
                lines = gff_file.readlines()
                new_lines = []
                for line in lines:
                    values = line.split("\t")
                    values[3] = str(int(values[3]) + int(start))
                    values[4] = str(int(values[4]) + int(start))
                    new_lines.append("\t".join(values))
                gff_file.close()
                gff_file = open(tmp_path + name + ".gff", "w")
                for line in new_lines:
                    gff_file.write(line)
                gff_file.close()
            except FileNotFoundError:
                pass
    region.close()
    output.close()

def searching_for_db(assembly_path):

    db_endings = ['.ndb', '.nhr', '.nin', '.nog', '.nos', '.not', '.nsq', '.ntf', '.nto']
    check = True
    for end in db_endings:
        if not any(File.endswith(end) for File in os.listdir(assembly_path)):
            check = False
    return check

def get_distance_biopython(file, matrix):
    """ Reads alignment file and returns distance matrix """
    #print(file)
    input_handle = open(file)
    aln = AlignIO.read(input_handle, 'fasta')
    try:
        calculator = DistanceCalculator(matrix)
        dm = calculator.get_distance(aln)
    except ValueError:
        #print('The amino acid U is scored as C during distance calculation for file %s'%(file))
        for record in aln:
            new_seq = record.seq.replace('U', 'C')
            record.seq = new_seq
        calculator = DistanceCalculator(matrix)
        dm = calculator.get_distance(aln)
    input_handle.close()
    return dm

def readFasta(fasta):
    path = Path(fasta)
    if path.exists() == False:
        print(str(path) + ' does not exists.')
        sys.exit()
    seq_records = SeqIO.parse(path, "fasta")
    return seq_records

def getSeedInfo(path):
    dic = {}
    seq_records = readFasta(path)
    for entry in seq_records:
        species = entry.id.split("|")[1]
        geneID = entry.id.split("|")[2]

        try:
            dic[species].append(geneID)
        except KeyError:
            dic[species] = [geneID]

    del seq_records
    return dic

def checkCoOrthologs(candidate_name, best_hit, ref, fdog_ref_species, candidatesOutFile, msaTool, matrix, dataPath, tmp_path, mode='silent'):
    ###########getting sequences and write all in one file to make msa #########
    name_file = candidate_name + ".co"
    output_file = tmp_path + name_file + '.fa'
    aln_file = tmp_path + name_file + '.aln'
    genome_dir_path = dataPath + '/searchTaxa_dir/%s/%s.fa'%(fdog_ref_species, fdog_ref_species)
    if not os.path.exists(genome_dir_path):
        genome_dir_path = dataPath + '/genome_dir/%s/%s.fa'%(fdog_ref_species, fdog_ref_species)
    #print(searchTool)

    out = open(output_file, "w")
    inSeq = SeqIO.to_dict((SeqIO.parse(open(genome_dir_path), 'fasta')))
    out.write(">" + best_hit + "\n")
    out.write(str(inSeq[best_hit].seq) + "\n")
    out.write(">" + ref + "\n")
    out.write(str(inSeq[ref].seq )+ "\n")
    #print(candidatesOutFile)
    candidates = readFasta(candidatesOutFile)
    for record in candidates:
        if candidate_name in record.id:
            out.write(">" + candidate_name + "\n")
            out.write(str(record.seq) + "\n")
            break

    out.close()

    if msaTool == "muscle":
        if align_fn.get_muscle_version(msaTool) == 'v3':
            cmd = "muscle -quiet -in " + output_file + "-out " + aln_file
        else:
            cmd = "muscle -align " + output_file +  " -output " + aln_file
        starting_subprocess(cmd, mode)
        if not os.path.exists(aln_file):
            print('Muscle failed with command: %s'%(cmd))
            print("Muscle failed for file %s. Making MSA with Mafft-linsi." % (candidate_name))
            cmd = 'mafft --maxiterate 1000 --localpair --anysymbol --quiet ' + output_file + ' > ' + aln_file
            starting_subprocess(cmd, mode)

    elif msaTool == "mafft-linsi":
        #print("mafft-linsi")
        cmd = 'mafft --maxiterate 1000 --localpair --anysymbol --quiet ' + output_file + ' > ' + aln_file
        starting_subprocess(cmd, mode)

    try:
        distances = get_distance_biopython(aln_file, matrix)
        distance_hit_query = distances[best_hit, candidate_name]
        distance_ref_hit = distances[best_hit, ref]
        #print(distances)
    except ValueError:
        pass
        #print("Failure in distance computation, Candidate  %s will be rejected" % candidate_name)
        return 0, "NaN", "NaN"

    #distance_hit_query = distances[best_hit, candidate_name]
    #distance_ref_hit = distances[best_hit, ref]

    if distance_ref_hit < distance_hit_query:
        #accepted
        return 1, distance_ref_hit, distance_hit_query

    else:
        #rejected
        return 0, distance_ref_hit, distance_hit_query

def backward_search(candidatesOutFile, fasta_path, strict, fdog_ref_species, evalue_cut_off, taxa, searchTool, checkCo, msaTool, matrix, dataPath, filter, tmp_path, mode):
    # the backward search uses the genes predicted from augustus and makes a blastp search
    #the blastp search is against all species that are part of the core_ortholog group if the option --strict was chosen or only against the ref taxa
    seedDic = getSeedInfo(fasta_path)
    #print(fasta_path)
    orthologs = []
    #print(seedDic)
    blast_dir_path = dataPath + "/coreTaxa_dir/"
    #print(blast_dir_path)
    if not os.path.exists(blast_dir_path):
        blast_dir_path = dataPath + "/blast_dir/"
    #print(blast_dir_path)
    if strict != True:
        seed = [fdog_ref_species]
        try:
            id_ref = seedDic[fdog_ref_species]
        except KeyError:
            #print("The fDOG reference species isn't part of the core ortholog group, ... exciting")
            return 0, seed
        if searchTool == "blast":
            cmd = "blastp -db " + blast_dir_path + fdog_ref_species + "/" + fdog_ref_species + " -outfmt '6 sseqid qseqid evalue' -max_target_seqs 10 -out " + tmp_path + "blast_" + fdog_ref_species + " -evalue " + str(evalue_cut_off) + " -query " + candidatesOutFile
            starting_subprocess(cmd, mode)
        else:
            print("diamonds are the girls best friends")
            ##### diamond call

        alg_file = open(tmp_path + "blast_" + fdog_ref_species, "r")
        lines = alg_file.readlines()
        alg_file.close()
        old_name = None
        min = 10
        for line in lines:
            id, gene, evalue = (line.replace("\n", "")).split("\t")
            gene_name = gene.split("|")[2]
            if gene_name != old_name:
                print("candidate:%s"%(gene_name)) if mode == "debug" else ""
                print("blast-hit:%s"%(id)) if mode == "debug" else ""
                min = float(evalue)
                if id in id_ref:
                    orthologs.append(gene)
                    print("\thitting\n") if mode == "debug" else ""
                else:
                    if checkCo == True:
                        for i in id_ref:
                            print("Best hit %s differs from reference sequence %s! Doing further checks\n"%(id, i)) if mode == "debug" else ""
                            co_orthologs_result, distance_ref_hit, distance_hit_query = checkCoOrthologs(gene_name, id, i, fdog_ref_species, candidatesOutFile, msaTool, matrix, dataPath, tmp_path)
                            if co_orthologs_result == 1:
                                print("\t Distance query - blast hit: %6.4f, Distance blast hit - reference: %6.4f\tAccepting\n"%(distance_hit_query, distance_ref_hit)) if mode == "debug" else ""
                                orthologs.append(gene)
                            elif co_orthologs_result == 0:
                                if distance_ref_hit != "NaN":
                                    print("\t Distance query - blast hit: %6.4f, Distance blast hit - reference: %6.4f\tRejecting\n"%(distance_hit_query, distance_ref_hit)) if mode == "debug" else ""
                    else:
                        print("\tnothitting\n") if mode == "debug" else ""
            elif (gene_name == old_name) and float(evalue) == min and gene_name not in orthologs:
                if id in id_ref:
                    orthologs.append(gene)
                    print("\thitting\n") if mode == "debug" else ""
                else:
                    if checkCo == True:
                        for i in id_ref:
                            print("Best hit %s differs from reference sequence %s! Doing further checks\n"%(id, i)) if mode == "debug" else ""
                            co_orthologs_result, distance_ref_hit, distance_hit_query = checkCoOrthologs(gene_name, id, i, fdog_ref_species, candidatesOutFile, msaTool, matrix, dataPath, tmp_path)
                            if co_orthologs_result == 1:
                                print("\t Distance query - blast hit: %6.4f, Distance blast hit - reference: %6.4f\tAccepting\n"%(distance_hit_query, distance_ref_hit)) if mode == "debug" else ""
                                orthologs.append(gene)
                            elif co_orthologs_result == 0:
                                print("\t Distance query - blast hit: %6.4f, Distance blast hit - reference: %6.4f\tRejecting\n"%(distance_hit_query, distance_ref_hit)) if mode == "debug" else ""
                    else:
                        print("\tnot hitting\n") if mode == "debug" else ""
            old_name = gene_name


        if orthologs == []:
            #print("No hit in the backward search, ...exciting")
            return 0, seed

    else:
        if taxa != []:
            seed = taxa
            try:
                i = seed.index(fdog_ref_species)
                seed.insert(0,seed.pop(i))
            except ValueError:
                seed.insert(0,fdog_ref_species)
            #print(seed)
            #print("with taxa list from user input")

        else:
            seed = []
            for key in seedDic:
                if key == fdog_ref_species:
                    seed.insert(0,key)
                else:
                    seed.append(key)

        orthologs = set({})

        for species in seed:
            #print("backward search in species %s\n" %species)
            orthologs_new = set({})
            try:
                id_ref = seedDic[species]
            except KeyError:
                #print("The species " + species + " isn't part of the core ortholog group, ... exciting")
                return 0, seed
            #cmd = "blastp -db " + blast_dir_path + fdog_ref_species + "/" + fdog_ref_species + " -outfmt '6 sseqid qseqid evalue' -max_target_seqs 10 -out " + tmp_path + "blast_" + fdog_ref_species + " -evalue " + str(evalue_cut_off) + " -query " + candidatesOutFile
            cmd = "blastp -db " + blast_dir_path + species + "/" + species + " -outfmt '6 sseqid qseqid evalue' -max_target_seqs 10 -out " + tmp_path + "/blast_" + species + " -evalue " + str(evalue_cut_off) + " -query " + candidatesOutFile
            results = starting_subprocess(cmd, mode)
            if results.returncode != 0:
                print("Blastp failed with the command: %s"%(results.args))
                sys.exit()
            alg_file = open(tmp_path + "/blast_" + species, "r")
            lines = alg_file.readlines()
            alg_file.close()
            old_name = None
            min = 10
            for line in lines:
                id, gene_name, evalue = (line.replace("\n", "")).split("\t")
                if gene_name != old_name:
                    min = float(evalue)
                    if id in id_ref:
                        orthologs_new.add(gene_name)

                elif (gene_name == old_name) and float(evalue) == min:
                    if id in id_ref:
                        orthologs_new.add(gene_name)

            #print(species)
            #print(orthologs_new)
            #print(orthologs)
            if species == fdog_ref_species:
                orthologs = orthologs_new
            else:
                orthologs = orthologs & orthologs_new
                if len(orthologs) == 0:
                    #print("No ortholog was found with option --strict")
                    return 0, seed

    #print(orthologs)
    orthologs = set(orthologs)
    return list(orthologs), seed

def addRef(output, core_fasta, species_list):
    #print(species_list)
    output_file = open(output, "a+")
    seq_records_core = readFasta(core_fasta)
    seq_records_core = list(seq_records_core)
    for species in species_list:
        for entry_core in seq_records_core:
            if species in entry_core.id:
                output_file.write(">" + entry_core.id + "\n")
                output_file.write(str(entry_core.seq) + "\n")
    output_file.close()

def addSeq(output, seq_list):
    output_file = open(output, "a+")

    for item in seq_list:
        #print(item)
        candidate_fasta = item[1]
        sequenceIds = item[0]
        if sequenceIds == 0 or sequenceIds == []:
            continue
        seq_records_candidate = readFasta(candidate_fasta)
        seq_records_candidate = list(seq_records_candidate)
        for entry_candidate in seq_records_candidate:
            if entry_candidate.id in sequenceIds:
                if entry_candidate.id == sequenceIds[0]:
                    output_file.write(">" + entry_candidate.id + "|1" + "\n")
                    output_file.write(str(entry_candidate.seq) + "\n")
                else:
                    output_file.write(">" + entry_candidate.id + "|0" + "\n")
                    output_file.write(str(entry_candidate.seq) + "\n")
    output_file.close()

def addSequences(sequenceIds, candidate_fasta, core_fasta, output, name, species_list, refBool, tmp_path):

    output_file = open(output, "a+")
    if refBool == False:
        seq_records_core = readFasta(core_fasta)
        seq_records_core = list(seq_records_core)
        for species in species_list:
            for entry_core in seq_records_core:
                if species in entry_core.id:
                    output_file.write(">" + entry_core.id + "\n")
                    output_file.write(str(entry_core.seq) + "\n")

    if sequenceIds != 0:
        seq_records_candidate = readFasta(candidate_fasta)
        seq_records_candidate = list(seq_records_candidate)
        for entry_candidate in seq_records_candidate:
            if entry_candidate.id in sequenceIds:
                if entry_candidate.id == sequenceIds[0]:
                    output_file.write(">" + entry_candidate.id + "|1" + "\n")
                    output_file.write(str(entry_candidate.seq) + "\n")
                else:
                    output_file.write(">" + entry_candidate.id + "|0" + "\n")
                    output_file.write(str(entry_candidate.seq) + "\n")
    output_file.close()
    return 0

def createFasInput(orthologsOutFile, mappingFile):
    with open(orthologsOutFile, "r") as f:
        fas_seed_id = (f.readline())[1:-1]
        #fas_seed_id = fas_seed_id.split("|")[0]

    mappingFile = open(mappingFile, "a+")

    seq_records = readFasta(orthologsOutFile)
    for seq in seq_records:
        ncbi_id = (seq.id.split("@"))[1]
        mappingFile.write(seq.id + "\t" + "ncbi" + ncbi_id + "\n")

    mappingFile.close()
    return fas_seed_id

def cleanup(tmp, tmp_path):
    if tmp == False:
        timeout = time.time() + 60*1
        while os.path.exists(tmp_path):
            shutil.rmtree(tmp_path, ignore_errors=True)
            if time.time() > timeout:
                print("tmp folder could not be removed!")
                break
    else:
        # clean up whole contigs
        for root, dirs, files in os.walk(tmp_path):
            for file in files:
                if file.endswith(".fasta"):
                    os.remove(os.path.join(root, file))

def getLocationFromGff(gff_file, name, tool):
    #print(name)
    if tool == 'metaeuk':
        gene_count = int(name.split('_')[-1:][0])
    else:
        gene_count = int(name.split('.')[-2].replace('g', '').split('_')[-1:][0])
    counter = 0
    with open(gff_file,'r') as gff:
        for line in gff:
            if line.startswith('#'):
                pass
            else:
                    contig, source, type, start, end, score, strand, phase, att = line.split('\t')
                    if type == 'gene':
                        counter += 1
                        if counter == gene_count:
                            position = [contig, int(start), int(end), strand]
                            #print(position)
                            return position

def checkOverlap(position, n=30):
    pairs = set()
    overlapping = set()
    keys = list(position.keys())
    index = 0
    for x in keys:
        index +=1
        for i in range(index,len(keys)):
            y = keys[i]
            if x != y:
                if position[y][0] == position[x][0]:
                    if position[y][3] == position[x][3]:
                        if position[x][1] < position[y][1] and position[y][1] <= position[x][2]:
                            len_overlap = position[x][2] - position[y][1]
                            if len_overlap >= n:
                                pairs.add((y,x))
                                overlapping.add(y)
                                overlapping.add(x)
                        elif position[x][1] == position[y][1]:
                            len_overlap = min(position[x][2],position[y][2])  - position[x][1]
                            if len_overlap >= n:
                                pairs.add((y,x))
                                overlapping.add(y)
                                overlapping.add(x)
                        elif position[x][2] == position[y][2]:
                            len_overlap = position[x][2] - max((position[x][2],position[y][2]))
                            if len_overlap >= n:
                                pairs.add((y,x))
                                overlapping.add(y)
                                overlapping.add(x)
                        elif position[y][1] < position[x][1] and position[x][1] <= position[y][2]:
                            len_overlap = position[y][2] - position[x][1]
                            if len_overlap >= n:
                                pairs.add((y,x))
                                overlapping.add(y)
                                overlapping.add(x)
    return pairs, overlapping

def coorthologs(candidate_names, tmp_path, candidatesFile, fasta, fdog_ref_species, msaTool, matrix, isoforms, gene_prediction, mode='silent'):
    if len(candidate_names) == 1:
        return candidate_names

    candidates = readFasta(candidatesFile)
    ref = readFasta(fasta)

    out = tmp_path + '/checkCoorthologs.fa'
    f = open(out,"w")

    aln_file = tmp_path + '/checkCoorthologs.aln'

    for record in ref:
        if fdog_ref_species in record.id:
            ref_id = record.id
            f.write(">" + record.id + "\n")
            f.write(str(record.seq) +  "\n")
            break

    already_written = []
    for record in candidates:
        for name in candidate_names:
            if name == record.id:
                if name not in already_written:
                    f.write(">" + record.id + "\n")
                    f.write(str(record.seq) + "\n")
                    already_written.append(name)
    f.close()

    if msaTool == "muscle":
        if align_fn.get_muscle_version(msaTool) == 'v3':
            cmd = "muscle -quiet -in %s -out %s" % (out, aln_file)
        else:
            cmd = "muscle -align %s -output %s" % (out, aln_file)
        starting_subprocess(cmd, mode)
        if not os.path.exists(aln_file):
            print('Muscle failed with command: %s' % (cmd))
            print("Muscle failed for file %s. Making MSA with Mafft-linsi." % (aln_file))
            cmd = 'mafft --maxiterate 1000 --localpair --anysymbol --quiet ' + out + ' > ' + aln_file
            starting_subprocess(cmd, mode)
    elif msaTool == "mafft-linsi":
        cmd = 'mafft --maxiterate 1000 --localpair --anysymbol --quiet %s > %s'% (out, aln_file)
        starting_subprocess(cmd, mode)

    distances = get_distance_biopython(aln_file, matrix)

    min_dist = 10
    min_name = None
    position = {}
    for name in candidate_names:
        distance = distances[ref_id , name]
        id = name.split('|')[2]
        if isoforms == False:
            gff_file = tmp_path + '/' + '_'.join(id.split('_')[0:-1]) + '.gff'
            position[name] = getLocationFromGff(gff_file, id, gene_prediction)
        if distance <= min_dist:
            min_dist = distance
            min_name = name

    checked = [min_name]
    pairs, overlapping = checkOverlap(position)
    #print(pairs, overlapping)
    tested = set()
    for name in candidate_names:
        if name == min_name:
            pass
        elif distances[min_name , name] <= distances[min_name , ref_id]:
            if isoforms == False and name in overlapping and name not in tested:
                for pair in pairs:
                    min_dist = 10
                    to_add = ''
                    if name in pair:
                        x,y = pair
                        tested.add(x)
                        tested.add(y)
                        distx = distances[x,ref_id]
                        disty = distances[y, ref_id]
                        if distx <= disty  and distx < min_dist:
                            to_add = x
                            min_dist = distx
                        elif disty <= distx  and disty < min_dist:
                            to_add = y
                            min_dist = disty
                if to_add != min_name:
                    checked.append(to_add)
            elif name in tested and isoforms == False:
                pass
            else:
                checked.append(name)
    #print(checked)
    return checked

def clean_fas(path, file_type):
    file = open(path, "r")
    lines = file.readlines()
    file.close()
    file = open(path,"w")

    for line in lines:
        if file_type == 'domains':
            long_id, remain = line.split("#")
            id = long_id.split("|")[0]
            new_line = id + "#" + remain
        else:
            long_id, remain = line.split("\t", 1)
            id = long_id.split("|")[0]
            new_line = id + "\t" + remain

        file.write(new_line)
    file.close()

def run_fas(cmd):
    #print(cmd)
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    while process.poll() is None:
        output = process.stdout.readline().decode().split('\n')
        error = process.stderr.readline().decode().split('\n')
        if error:
            cmd_out = ''
            for line in error:
                line.strip()
                if 'error' in line or 'Error' in line:
                    for i in cmd:
                        if '|' in i:
                            cmd_out += " '"+ str(i) + "'"
                        else:
                            cmd_out += " " + str(i)
                    print ("Error running FAS with %s"%(cmd_out))
                    process.terminate()
                    sys.exit()
    return output

def ortholog_search_tblastn(args):
    (asName, out, assemblyDir, consensus_path, augustus_ref_species, group, length_extension, average_intron_length, evalue, strict, fdog_ref_species, msaTool, matrix, dataPath, filter, mode, fasta_path, profile_path, taxa, searchTool, checkCoorthologs, gene_prediction, metaeuk_db, isoforms) = args
    output = []
    asNamePath = asName.replace('@', '_')
    cmd = 'mkdir ' + out + '/tmp/' + asNamePath
    starting_subprocess(cmd, 'silent')
    tmp_path = out + "tmp/" + asNamePath + "/"
    candidatesOutFile = tmp_path + group + ".candidates.fa"

    output.append("Searching in species " + asName + "\n")
    assembly_path = assemblyDir + "/" + asName + "/" + asName + ".fa"
    db_path = assemblyDir + "/" + asName + "/blast_dir/" + asName + ".fa"
    blast_dir_path = assemblyDir + "/" + asName + "/blast_dir/"
    if not os.path.exists(blast_dir_path):
        cmd = 'mkdir ' + blast_dir_path
        starting_subprocess(cmd, 'silent')
    db_check = searching_for_db(blast_dir_path)

    if db_check == 0:
        cmd = 'makeblastdb -in ' + assembly_path + ' -dbtype nucl -parse_seqids -out ' + db_path
        starting_subprocess(cmd, mode)

    #makes a tBLASTn search against database
    #codon table argument [-db_gencode int_value], table available ftp://ftp.ncbi.nih.gov/entrez/misc/data/gc.prt
    cmd = 'tblastn -db ' + db_path + ' -query ' + consensus_path + ' -outfmt "6 sseqid sstart send evalue qstart qend score " -evalue ' + str(evalue) + ' -out ' + tmp_path + '/blast_results.out'
    time_tblastn_start = time.time()
    exit_code = starting_subprocess(cmd, mode, 3600)
    time_tblastn_end = time.time()
    time_tblastn = time_tblastn_end - time_tblastn_start
    if exit_code == 1:
        output.append("The tblastn search takes too long for species %s. Skipping species ..." % asName)
        return [], candidatesOutFile, output

    output.append("Time tblastn %s in species %s" % (str(time_tblastn), asName))

    regions, number_regions = candidate_regions(average_intron_length, evalue, tmp_path)
    if regions == 0:
        #no candidat region are available, no ortholog can be found
        output.append("No candidate region found for species %s!\n" % asName)
        return [], candidatesOutFile, output

    else:
        output.append(str(number_regions) + " candiate region(s) were found for species %s.\n" % asName)
        extract_seq(regions, db_path, tmp_path, mode)

    if gene_prediction == "augustus":
        ############### make Augustus PPX search ###################################
        time_augustus_start = time.time()
        augustus_ppx(regions, candidatesOutFile, length_extension, profile_path, augustus_ref_species, asName, group, tmp_path, mode)
        time_augustus_end = time.time()
        time_augustus = time_augustus_end - time_augustus_start
        output.append("Time augustus: %s species %s \n" % (str(time_augustus), asName))
    else:
        time_metaeuk_start = time.time()
        if metaeuk_db == '':
            db = fasta_path
        else:
            db = metaeuk_db
        metaeuk_single(regions, candidatesOutFile, length_extension, asName, group, tmp_path, mode, db)
        time_metaeuk_end = time.time()
        time_metaeuk = time_metaeuk_end - time_metaeuk_start
        output.append("Time metaeuk: %s species %s \n" % (str(time_metaeuk), asName))

    ################# backward search to filter for orthologs###################
    if int(os.path.getsize(candidatesOutFile)) <= 0:
        #print("No genes found at candidate regions\n")
        return [], candidatesOutFile, output

    reciprocal_sequences, taxa = backward_search(candidatesOutFile, fasta_path, strict, fdog_ref_species, evalue, taxa, searchTool, checkCoorthologs, msaTool, matrix, dataPath, filter, tmp_path, mode)

    if reciprocal_sequences == 0:
        if regions != 0:
            output.append("No ortholog fulfilled the reciprocity criteria for species %s.\n" % asName)
        return [], candidatesOutFile, output
    else:
        reciprocal_sequences = coorthologs(reciprocal_sequences, tmp_path, candidatesOutFile, fasta_path, fdog_ref_species, msaTool, matrix, isoforms, gene_prediction)

    return reciprocal_sequences, candidatesOutFile, output

def blockProfiles(core_path, group, mode, out, msaTool):

    ######################## paths ################################
    msa_path = core_path + "/" + group +"/"+ group + ".aln"
    if not os.path.exists(msa_path):
        fasta_path = core_path + "/" + group +"/"+ group + ".fa"
        check_path(fasta_path)
        if msaTool == "muscle":
            if align_fn.get_muscle_version(msaTool) == 'v3':
                cmd = "muscle -quiet -in " + fasta_path + " -out " + msa_path
            else:
                cmd = "muscle -quiet -align " + fasta_path + " -output " + msa_path
        elif msaTool == "mafft-linsi":
            cmd = 'mafft --maxiterate 1000 --localpair --anysymbol --quiet ' + fasta_path + ' > ' + msa_path
        starting_subprocess(cmd, mode)

    profile_path = out + "/tmp/" + group + ".prfl"

    ######################## block profile #####################################

    print("Building a block profile ...", flush=True)
    cmd = 'msa2prfl.pl ' + msa_path + ' --setname=' + group + ' >' + profile_path
    starting_subprocess(cmd, 'silent')

    if int(os.path.getsize(profile_path)) > 0:
        print("\t ...finished \n")
    else:
        print("Building block profiles failed. Using prepareAlign to convert alignment\n")
        new_path = core_path + group +"/"+ group + "_new.aln"
        cmd = 'prepareAlign < ' + msa_path + ' > ' + new_path
        starting_subprocess(cmd, mode)
        cmd = 'msa2prfl.pl ' + new_path + ' --setname=' + group + ' >' + profile_path
        starting_subprocess(cmd, 'silent')
        print(" \t ...finished \n", flush=True)

    return profile_path

def consensusSequence(core_path, group, mode, out):

    ######################## paths ################################
    hmm_path = core_path + "/" + group +"/hmm_dir/"+ group + ".hmm"
    check_path(hmm_path)
    consensus_path = out + "/tmp/" + group + ".con"

    ######################## consensus sequence ################################
    #make a majority-rule consensus sequence with the tool hmmemit from hmmer
    print("Building a consensus sequence")
    cmd = 'hmmemit -c -o' + consensus_path + ' ' + hmm_path
    starting_subprocess(cmd, mode)
    print("\t ...finished\n")

    return consensus_path

def createGff(ortholog_sequences, out_folder, tool):
    #print(ortholog_sequences)
    #print(out_folder)
    gff_folder = out_folder + "/gff/"
    os.system('mkdir %s >/dev/null 2>&1' %(gff_folder))
    types_set = set(['gene', 'CDS', 'transcript', 'mRNA', 'exon'])
    for s in ortholog_sequences:
        genes = s[0]
        #print(genes)
        data = []
        if genes != []:
            gff_file_sp = gff_folder + '/' + genes[0].split('|')[1] + '.gff'
            for gene in genes:
                if gene == '':
                    continue
                #print(gene.split('|'))
                group, species, gene = gene.split('|')
                #print(group, species, gene)
                region = '_'.join(gene.split('_')[0:-1])
                if tool == 'metaeuk':
                    gene_count = int(gene.split('_')[-1:][0])
                else:
                    gene_count = int(gene.split('.')[-2].replace('g', '').split('_')[-1:][0])
                #print(region, gene_count)
                gff_file_gene = "%s/tmp/%s/%s.gff" %(out_folder, species.replace('@', '_'), region)
                #print(gff_file_gene)
                with open(gff_file_gene, 'r') as gff:
                    counter = 0
                    for line in gff:
                        if line.startswith('#'):
                            pass
                        else:
                            line=line.rstrip()
                            contig, source, type, start, end, score, strand, phase, att = line.split('\t')
                            if type == 'gene':
                                counter += 1
                            if counter == gene_count:
                                if source == 'AUGUSTUS':
                                    att = att.replace('g' + str(gene_count), group + '_' + '_'.join(gene.split('.')[:-1]))
                                    att = att.replace('"', '')
                                    if type not in types_set:
                                        continue
                                elif source == 'MetaEuk':
                                    #att = 'ID=' + group + '_' +  gene + '; ' +  att
                                    if type == 'gene':
                                        att_entries = att.split(';')
                                        for x in att_entries:
                                            if x.startswith('Target_ID='):
                                                target = x
                                            elif x.startswith('TCS_ID='):
                                                parent_prefix = x.replace('TCS_ID=', '')
                                    att = att.replace('TCS_ID=', 'ID=')
                                    att = att.replace(parent_prefix, group + '_' + gene)
                                    att = att.replace(target + ';', '')
                                    phase = 0
                                data.append([contig, source, type, int(start), int(end), score, strand, phase, att])
        else:
            continue

        df = pd.DataFrame(data, columns=['contig', 'source', 'type', 'start', 'end', 'score', 'starnd', 'phase', 'att'])
        #print(df)
        df.sort_values(by=['contig', 'start'])
        df.to_csv(gff_file_sp,sep='\t' , index=False, header=None)

def getAugustusRefSpec(mapping_augustus):
    dict = {}
    with open(mapping_augustus,'r') as file:
        for line in file:
            line = line.rstrip()
            assembly, id = line.split('\t')
            dict[assembly] = id
    return dict

def main():

    #################### handle user input #####################################

    start = time.time()
    version = '0.1.5.2'
    ################### initialize parser ######################################
    parser = argparse.ArgumentParser(description='You are running fdog.assembly version ' + str(version) + '.')
    parser.add_argument('--version', action='version', version=str(version))
    ################## required arguments ######################################
    required = parser.add_argument_group('Required arguments')
    required.add_argument('--gene', help='Core_ortholog group name. Folder inlcuding the fasta file, hmm file and aln file has to be located in core_orthologs/',
                            action='store', default='', required=True)
    required.add_argument('--refSpec', help='Reference taxon/taxa for fDOG.', action='store', nargs="+", default='', required=True)
    ################## optional arguments ######################################
    optional = parser.add_argument_group('Optional arguments')
    optional.add_argument('--avIntron', help='average intron length of the assembly species in bp (default: 50000)',action='store', default=50000, type=int)
    optional.add_argument('--lengthExtension', help='length extension of the candidate regions in bp (default:20000)', action='store', default=20000, type=int)
    optional.add_argument('--assemblyPath', help='Path for the assembly directory, (default dataPath)', action='store', default='')
    optional.add_argument('--tmp', help='tmp files will not be deleted', action='store_true', default = False)
    optional.add_argument('--out', help='Output directory', action='store', default='')
    optional.add_argument('--dataPath', help='fDOG data directory containing searchTaxa_dir, coreTaxa_dir and annotation_dir', action='store', default='')
    optional.add_argument('--coregroupPath', help='core_ortholog directory containing ortholog groups of gene of interest', action='store', default='')
    #optional.add_argument('--searchTool', help='Choose between blast and diamond as alignment search tool(default:blast)', action='store', choices=['blast', 'diamond'], default='blast')
    optional.add_argument('--evalBlast', help='E-value cut-off for the Blast search. (default: 0.00001)', action='store', default=0.00001, type=float)
    optional.add_argument('--strict', help='An ortholog is only then accepted when the reciprocity is fulfilled for each sequence in the core set', action='store_true', default=False)
    optional.add_argument('--msaTool', help='Choose between mafft-linsi or muscle for the multiple sequence alignment. (default:muscle)', choices=['mafft-linsi', 'muscle'], action='store', default='muscle')
    optional.add_argument('--checkCoorthologsOff', help='During the final ortholog search, accept an ortholog also when its best hit in the reverse search is not the core ortholog itself, but a co-ortholog of it', action='store_false', default=True)
    optional.add_argument('--scoringmatrix', help='Choose a scoring matrix for the distance criteria used by the option --checkCoorthologsRef. (default: blosum62)', choices=['identity', 'blastn', 'trans', 'benner6', 'benner22', 'benner74', 'blosum100', 'blosum30', 'blosum35', 'blosum40', 'blosum45', 'blosum50', 'blosum55', 'blosum60', 'blosum62', 'blosum65', 'blosum70', 'blosum75', 'blosum80', 'blosum85', 'blosum90', 'blosum95', 'feng', 'fitch', 'genetic', 'gonnet', 'grant', 'ident', 'johnson', 'levin', 'mclach', 'miyata', 'nwsgappep', 'pam120', 'pam180', 'pam250', 'pam30', 'pam300', 'pam60', 'pam90', 'rao', 'risler', 'structure'], action='store', default='blosum62')
    optional.add_argument('--coreTaxa', help='List of core taxa used during --strict', action='store', nargs="+", default=[])
    #optional.add_argument('--filter', help='Switch the low complexity filter for the blast search on.', action='store', default='no')
    optional.add_argument('--fasoff', help='Turn off FAS support', action='store_true', default=False)
    optional.add_argument('--pathFile', help='Config file contains paths to data folder (in yaml format)', action='store', default='')
    optional.add_argument('--searchTaxa', help='List of Taxa to search in, (default: all species located in assembly_dir)', action='store', nargs="+", default=[])
    optional.add_argument('--debug', help='Stdout and Stderr from fdog.assembly and every used tool will be printed, caution: using --parallel can result in messy output', action='store_true', default=False)
    optional.add_argument('--force', help='Overwrite existing output files', action='store_true', default=False)
    optional.add_argument('--append', help='Append the output to existing output files, caution: reference species must be identical', action='store_true', default=False)
    optional.add_argument('--parallel', help= 'The ortholog search of multiple species will be done in parallel', action='store_true', default=False)
    optional.add_argument('--augustus', help= 'Gene prediction is done by using the tool Augustus PPX', action='store_true', default=False)
    optional.add_argument('--augustusRefSpec', help='Augustus reference species identifier (use command: augustus --species=help to get precomputed augustus gene models)', action='store', default='')
    optional.add_argument('--augustusRefSpecFile', help='Mapping file tab seperated containing Assembly Names and augustus reference species that should be used', action='store', default='')
    optional.add_argument('--metaeukDb', help='Path to MetaEuk reference database', action='store', default='')
    optional.add_argument('--isoforms', help='All Isoforms of a gene passing the ortholog verification will be included in the output', action='store_true', default=False)
    optional.add_argument('--gff', help='GFF files will be included in output', action='store_true', default=False)
    args = parser.parse_args()

    # required
    group = args.gene
    fdog_ref_species = args.refSpec
    #paths user input
    assemblyDir = args.assemblyPath
    dataPath = args.dataPath
    core_path = args.coregroupPath
    out = args.out
    pathFile = args.pathFile
    #I/O
    tmp = args.tmp
    strict = args.strict
    checkCoorthologs = args.checkCoorthologsOff
    # print(checkCoorthologs)
    #others
    average_intron_length = args.avIntron
    length_extension = args.lengthExtension
    #searchTool = args.searchTool
    searchTool = 'blast'
    evalue = args.evalBlast
    msaTool = args.msaTool
    matrix = args.scoringmatrix
    taxa = args.coreTaxa
    fasoff = args.fasoff
    searchTaxa = args.searchTaxa
    debug = args.debug
    force = args.force
    append = args.append
    parallel = args.parallel
    augustus_ref_species = args.augustusRefSpec
    mapping_augustus = args.augustusRefSpecFile
    metaeuk_db = args.metaeukDb
    isoforms = args.isoforms
    gff = args.gff

    #gene prediction tool
    augustus = args.augustus
    if augustus == True:
        if augustus_ref_species == '' and mapping_augustus == '':
            print("Augustus reference species is required when using Augustus as gene prediction tool")
            return 1
        gene_prediction = "augustus"
        if mapping_augustus != '':
            check_path(mapping_augustus)
            aug_ref_dict = getAugustusRefSpec(mapping_augustus)
    else:
        gene_prediction = "metaeuk"
        if metaeuk_db == '':
            print("MetaEuk DB is required when using MetaEuk as gene prediction tool")
            return 1

    # output modes
    if debug == True:
        mode = 'debug'
    else:
        mode = 'normal'

    #checking paths
    if dataPath == '':
        fdogPath = os.path.realpath(__file__).replace('/fDOGassembly.py','')
        configFile = fdogPath + '/bin/pathconfig.yml'
        if not os.path.exists(configFile):
            sys.exit(
                f'No pathconfig.yml found at {configFile}. Please run fdog.setup '
                + '(https://github.com/BIONF/fDOG/wiki/Installation#setup-fdog).')
        if pathFile:
            configFile = os.path.abspath(pathFile)
        cfg = general_fn.load_config(configFile)
        try:
            dataPath = cfg['datapath']
        except:
            dataPath = os.getcwd()

    if out == '':
        out = os.getcwd()
    else:
        if out[-1] != "/":
            out = out + "/"
        if not os.path.exists(out):
            os.mkdir(out)
        check_path(out)
    out = os.path.abspath(out)

    if os.path.exists(out + '/' + group):
        if append != True and force != True:
            print("Output folder for group " + group + " exists already. Please choose --force or --append.")
            sys.exit()
        elif force == True:
            shutil.rmtree(out + '/' + group, ignore_errors=True)
            os.system('mkdir ' + out + '/' + group + ' >/dev/null 2>&1')
            out = out + '/' + group + '/'
        elif append == True:
            out = out + '/' + group + '/'
    else:
        os.system('mkdir ' + out + '/' + group + ' >/dev/null 2>&1')
        out = out + '/' + group + '/'

    if core_path == '':
        core_path = out + '/core_orthologs/'
        if check_path(core_path, False) == 1:
            core_path = dataPath + '/core_orthologs/'
    else:
        if not core_path.endswith('/'):
            core_path = core_path + '/'
        check_path(core_path)

    if assemblyDir == '':
        assemblyDir = dataPath + '/assembly_dir/'
    check_path(assemblyDir)

    if metaeuk_db != '':
        check_path(metaeuk_db)

    ################## How to handle std output and std error ##################

    if mode == 'silent':
        sys.stderr = False
        sys.stdout = False
    else:
        pass

    ########################### other variables ################################
    if searchTaxa == []:
        assembly_names = os.listdir(assemblyDir)
    else:
        if len(searchTaxa) > 1:
            assembly_names = os.listdir(assemblyDir)
            for Taxon in searchTaxa:
                if Taxon not in assembly_names:
                    print("Taxon %s is not in the assembly_dir" % Taxon)
                    sys.exit()
            assembly_names = searchTaxa
        else:
            if searchTaxa[0] in os.listdir(assemblyDir):
                assembly_names = searchTaxa
            elif os.path.isfile(searchTaxa[0]):
                with open(searchTaxa[0]) as file:
                    lines = file.readlines()
                    assembly_names = [line.rstrip() for line in lines]
            else:
                print("Input %s for search Taxa is not in the assembly_dir or an existing file" % searchTaxa[0])

    ################################# paths ####################################

    fasta_path = core_path + "/" + group +"/"+ group + ".fa"
    check_path(fasta_path)
    tmp_folder = out + "/tmp"

    ########### is/are fDOG reference species part of ortholog group? ##########

    fdog_ref_species = check_ref_spec(fdog_ref_species, fasta_path)

    ###################### create tmp folder ###################################

    cmd = 'mkdir ' + out + '/tmp'
    starting_subprocess(cmd, 'silent')

    print("Gene: " + group, flush=True)
    print("fDOG reference species: " + fdog_ref_species + " \n",flush=True)

    ###################### preparations ########################################

    if augustus == True:
        group_computation_time_start = time.time()
        consensus_path = core_path + '/' + group + '/' + group + '.con'
        if check_path(consensus_path, exit=False) == 1:
            consensus_path = consensusSequence(core_path, group, mode, out)
        print(consensus_path)
        profile_path = core_path + '/' + group + '/' + group + '.prfl'
        if check_path(profile_path, exit=False) == 1:
            profile_path = blockProfiles(core_path, group, mode, out, msaTool)
        print(profile_path)
        group_computation_time_end = time.time()
        time_group = group_computation_time_end - group_computation_time_start
    else:
        #print("test")
        profile_path = ""
        group_computation_time_start = time.time()
        consensus_path = core_path + '/' + group + '.con'
        if check_path(consensus_path, exit=False) == 1:
            consensus_path = consensusSequence(core_path, group, mode, out)
        #concatinade core_group sequences if metaeuk should be run without tblastn
        group_computation_time_end = time.time()
        time_group = group_computation_time_end - group_computation_time_start


    ###################### ortholog search #####################################

    ortholog_sequences = []
    time_ortholog_start = time.time()

    if parallel == True:
        ##################### parallel computation #############################
        calls = []
        cpus = mp.cpu_count()
        pool = mp.Pool(cpus)
        for asName in assembly_names:
            if mapping_augustus == '':
                calls.append([asName, out, assemblyDir, consensus_path, augustus_ref_species, group, length_extension, average_intron_length, evalue, strict, fdog_ref_species, msaTool, matrix, dataPath, filter, mode, fasta_path, profile_path, taxa, searchTool, checkCoorthologs, gene_prediction, metaeuk_db, isoforms])
            else:
                try:
                    calls.append([asName, out, assemblyDir, consensus_path, aug_ref_dict[asName], group, length_extension, average_intron_length, evalue, strict, fdog_ref_species, msaTool, matrix, dataPath, filter, mode, fasta_path, profile_path, taxa, searchTool, checkCoorthologs, gene_prediction, metaeuk_db, isoforms])
                except KeyError:
                    print("%s is not included in Augustus reference species mapping file. %s will be skipped" %(asName, asName))

        print("Searching for orthologs ...", flush=True)
        for i in tqdm(pool.imap_unordered(ortholog_search_tblastn, calls),total=len(calls)):
            ortholog_sequences.append([i[0], i[1]])
            if mode == 'debug':
                for k in i[2]:
                    print(k)
        print("\t ...finished \n", flush=True)

    else:
        ###################### computation species wise ################
        for asName in tqdm(assembly_names):
            if mapping_augustus == '':
                args = [asName, out, assemblyDir, consensus_path, augustus_ref_species, group, length_extension, average_intron_length, evalue, strict, fdog_ref_species, msaTool, matrix, dataPath, filter, mode, fasta_path, profile_path, taxa, searchTool, checkCoorthologs, gene_prediction, metaeuk_db, isoforms]
            else:
                try:
                    args = [asName, out, assemblyDir, consensus_path, augustus_ref_species, group, length_extension, average_intron_length, evalue, strict, fdog_ref_species, msaTool, matrix, dataPath, filter, mode, fasta_path, profile_path, taxa, searchTool, checkCoorthologs, gene_prediction, metaeuk_db, isoforms]
                except KeyError:
                    print("%s is not included in Augustus reference species mapping file. %s will be skipped" % (asName, asName))
            reciprocal_sequences, candidatesOutFile, output_ortholog_search = ortholog_search_tblastn(args)
            ortholog_sequences.append([reciprocal_sequences, candidatesOutFile])
            if mode == 'debug':
                for k in output_ortholog_search:
                    print(k)

    time_ortholog_end = time.time()
    time_ortholog = time_ortholog_end - time_ortholog_start

    ################## preparing output ########################################
    orthologsOutFile = out + "/" + group + "_og.fa"

    if taxa == []:
        taxa = [fdog_ref_species]
    if append == True:
        addSeq(orthologsOutFile, ortholog_sequences)
    else:
        addRef(orthologsOutFile, fasta_path, taxa)
        addSeq(orthologsOutFile, ortholog_sequences)

    if gff == True:
        createGff(ortholog_sequences, out, gene_prediction)
    mappingFile = out + "/tmp/" + group + ".mapping.txt"

    if fasoff == False:
        fas = time.time()
        print("Calculating FAS scores ...", flush=True)

        tmp_path = out + '/tmp/'
        fas_seed_id = createFasInput(orthologsOutFile, mappingFile)
        cmd = ['fas.run', '--seed', fasta_path , '--query' , orthologsOutFile , '--annotation_dir' , tmp_path + 'anno_dir' ,'--bidirectional', '--tsv', '--phyloprofile', mappingFile, '--seed_id', fas_seed_id, '--out_dir', out, '--out_name', group]
        # print(cmd)
        fas_out = run_fas(cmd)
        clean_fas(out + group + "_forward.domains", 'domains')
        clean_fas(out + group + "_reverse.domains", 'domains')
        clean_fas(out + group + ".phyloprofile", 'phyloprofile')
        print("\t ...finished \n", flush=True)
        end = time.time()
        time_fas = end - fas
    else:
        end = time.time()
        time_fas = 0

    ################# remove tmp folder ########################################

    print(
        f"fDOG-Assembly finished completely in {round(end-start,2)}s ("
        f" Group preparation: {round(time_group,2)}s \t"
        f"Ortholog search: {round(time_ortholog,2)}s \t"
        f"FAS: {round(time_fas,2)}s)"
    )
    print(f"Outputs are saved at {out}")
    sys.stdout = sys.__stdout__
    cleanup(tmp, tmp_folder)

if __name__ == '__main__':
    main()
